import itertools
import os

import vlcp.service.sdn.ofpportmanager as ofpportmanager
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ioprocessing as iop
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.server.module import depend, call_api
from vlcp.config.config import defaultconfig
from vlcp.event.runnable import RoutineContainer
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.ethernet import mac_addr_bytes, ip4_addr_bytes,ip4_icmp_payload,\
    ethernet_l7, ip4_packet_l7, ip4_payload,ICMP_ECHOREPLY,icmp_bestparse,icmp_echo,\
    ip_frag
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent, OpenflowAsyncMessageEvent
from vlcp.utils.networkmodel import SubNet,RouterPort
from namedstruct.stdprim import uint16
from vlcp.event.event import M_

class ICMPResponderUpdater(FlowUpdater):
    def __init__(self,connection,parent):
        super(ICMPResponderUpdater,self).__init__(connection,(),('icmpresponderupdate',connection),parent._logger)
        self.parent = parent
        self._lastlognets = ()
        self._lastlogports = ()
        self._lastsubnetsinfo = dict()
        self._orig_initialkeys = ()

    async def main(self):
        try:
            self.subroutine(self._update_handler(),True,"update_handler_routine")

            # use controller to reply icmp ping ,so start routine handler packet in
            if not self.parent.prepush:
                self.subroutine(self._icmp_packetin_handler(),True,"icmp_packetin_handler_routine")

            await FlowUpdater.main(self)
        finally:
            if hasattr(self,"update_handler_routine"):
                self.update_handler_routine.close()

            if hasattr(self,"icmp_packetin_handler_routine"):
                self.icmp_packetin_handler_routine.close()

    async def _icmp_packetin_handler(self):
        conn = self._connection
        ofdef = self._connection.openflowdef
        l3input = self.parent._gettableindex("l3input",self._connection.protocol.vhost)
        
        transactid = uint16.create(os.urandom(2)) 

        async def send_packet_out(portid,packet):
            await self.execute_commands(conn,
                        [
                            ofdef.ofp_packet_out(
                                buffer_id = ofdef.OFP_NO_BUFFER,
                                in_port = ofdef.OFPP_CONTROLLER,
                                actions = [
                                    ofdef.ofp_action_output(port = portid,
                                                            max_len = ofdef.OFPCML_NO_BUFFER
                                                            )
                                ],
                                data = packet._tobytes()
                            )
                        ])

        icmp_packetin_matcher = OpenflowAsyncMessageEvent.createMatcher(ofdef.OFPT_PACKET_IN,None,None,l3input,2,
                                                            self._connection,self._connection.connmark)

        while True:
            ev = await icmp_packetin_matcher
            msg = ev.message
            inport = ofdef.ofp_port_no.create(ofdef.get_oxm(msg.match.oxm_fields,ofdef.OXM_OF_IN_PORT))

            # it must be icmp packet ...
            icmp_packet = ethernet_l7.create(msg.data)
            
            if ip_frag(icmp_packet) != 0:
                # ignore fragmented packets
                continue
            
            transactid = (transactid + 1) & 0xffff
            reply_packet = ip4_packet_l7((ip4_payload,ip4_icmp_payload),
                                         (icmp_bestparse, icmp_echo),
                                        dl_src = icmp_packet.dl_dst,
                                        dl_dst = icmp_packet.dl_src,
                                        ip_src = icmp_packet.ip_dst,
                                        ip_dst = icmp_packet.ip_src,
                                        frag_off = 0,
                                        ttl = 128,
                                        identifier = transactid,
                                        icmp_type = ICMP_ECHOREPLY,
                                        icmp_code = icmp_packet.icmp_code,
                                        icmp_id = icmp_packet.icmp_id,
                                        icmp_seq = icmp_packet.icmp_seq,
                                        data = icmp_packet.data
                                        )
           
            self.subroutine(send_packet_out(inport,reply_packet))
    
    async def _update_handler(self):

        # when lgport,lgnet,phyport,phynet object change , receive this event from ioprocessing module
        dataobjectchange = iop.DataObjectChanged.createMatcher(None,None,self._connection)

        while True:
            ev = await dataobjectchange

            # save to instance attr ,  us in other method
            self._lastlogports,_,self._lastlognets,_ = ev.current
            self._update_walk()

    def _walk_lgport(self,key,value,walk,save):

        if value is not None:
            save(key)
            if hasattr(value,'subnet'):
                try:
                    subnetobj = walk(value.subnet.getkey())
                except KeyError:
                    pass
                else:
                    save(value.subnet.getkey())
                    if subnetobj is not None and hasattr(subnetobj,"router"):
                        try:
                            _ = walk(subnetobj.router.getkey())
                        except KeyError:
                            pass
                        else:
                            save(subnetobj.router.getkey())

    def _walk_lgnet(self,key,value,walk,save):
        save(key)
        # if value is None, also save its key
        # means watch key, when created , we will recv event

    def _update_walk(self):
        lgportkeys = [p.getkey() for p,_ in self._lastlogports]
        lgnetkeys = [p.getkey() for p,_ in self._lastlognets]

        self._initialkeys = lgportkeys + lgnetkeys
        self._orig_initialkeys = lgportkeys + lgnetkeys
        
        self._walkerdict = dict(itertools.chain(((p,self._walk_lgport) for p in lgportkeys),
                                                ((n,self._walk_lgnet) for n in lgnetkeys)))
        
        self.subroutine(self.restart_walk(),False)
    
    def reset_initialkeys(self,keys,values):
        # walk map  logicalport --> subnet ---> routerport
        # we get subnet object, add keys to initialkeys, 
        # when subnet update, it will restart walk ,, after we will get new routerport
        
        subnetkeys = [k for k,v in zip(keys,values) if v is not None and not v.isdeleted() and
                      v.isinstance(SubNet)]

        self._initialkeys = tuple(itertools.chain(self._orig_initialkeys,subnetkeys))
    
    async def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        try:
            allobjects = set(o for o in self._savedresult if o is not None and not o.isdeleted())

            lastsubnetsinfo = self._lastsubnetsinfo
            currentlognetsinfo = dict((n,id) for n,id in self._lastlognets if n in allobjects)
            currentrouterportsinfo = dict((o.subnet,o) for o in allobjects
                                            if o.isinstance(RouterPort))
            currentsubnetsinfo = dict((o,(getattr(currentrouterportsinfo[o],"ip_address",getattr(o,"gateway",None)),
                                          self.parent.inroutermac,o.network.id,currentlognetsinfo[o.network]))
                                        for o in allobjects if o.isinstance(SubNet)
                                            and hasattr(o,"router") and o in currentrouterportsinfo
                                            and o.network in currentlognetsinfo
                                            and (hasattr(currentrouterportsinfo[o],"ip_address")
                                                or hasattr(o,"gateway"))
                                            and ( not hasattr(o,"isexternal") or o.isexternal == False))
            self._lastsubnetsinfo = currentsubnetsinfo

            ofdef = connection.openflowdef
            vhost = connection.protocol.vhost
            l3input = self.parent._gettableindex("l3input",vhost)

            cmds = []

            if connection.protocol.disablenxext:
                def match_network(nid):
                    return ofdef.create_oxm(ofdef.OXM_OF_METADATA_W, (nid & 0xffff) << 32,
                                            b'\x00\x00\xff\xff\x00\x00\x00\x00')
            else:
                def match_network(nid):
                    return ofdef.create_oxm(ofdef.NXM_NX_REG4, nid)

            # prepush or not ,, it is same , so ..
            def _deleteicmpflows(ipaddress, macaddress, networkid):
                    return [
                        ofdef.ofp_flow_mod(
                            cookie = 0x2,
                            cookie_mask = 0xffffffffffffffff,
                            table_id = l3input,
                            command = ofdef.OFPFC_DELETE,
                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                            buffer_id = ofdef.OFP_NO_BUFFER,
                            out_port = ofdef.OFPP_ANY,
                            out_group = ofdef.OFPG_ANY,
                            match = ofdef.ofp_match_oxm(
                                oxm_fields = [
                                    ofdef.create_oxm(ofdef.NXM_NX_REG4,networkid),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST,mac_addr_bytes(macaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,ofdef.ETHERTYPE_IP),
                                    ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,ip4_addr_bytes(ipaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO,ofdef.IPPROTO_ICMP),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_TYPE,8),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_CODE,0)
                                ]
                            )
                        )
                    ]

            if not self.parent.prepush:
                def _createicmpflows(ipaddress, macaddress, networkid):
                    return [
                        ofdef.ofp_flow_mod(
                            cookie = 0x2,
                            cookie_mask = 0xffffffffffffffff,
                            table_id = l3input,
                            command = ofdef.OFPFC_ADD,
                            # icmp to router matcher same as ip forward to router
                            # so priority + 1
                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                            buffer_id = ofdef.OFP_NO_BUFFER,
                            out_port = ofdef.OFPP_ANY,
                            out_group = ofdef.OFPG_ANY,
                            match = ofdef.ofp_match_oxm(
                                oxm_fields = [
                                    match_network(networkid),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST,mac_addr_bytes(macaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,ofdef.ETHERTYPE_IP),
                                    ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,ip4_addr_bytes(ipaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO,ofdef.IPPROTO_ICMP),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_TYPE,8),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_CODE,0)
                                ]
                            ),
                            instructions = [
                                ofdef.ofp_instruction_actions(
                                    actions = [
                                        ofdef.ofp_action_output(
                                            port = ofdef.OFPP_CONTROLLER,
                                            max_len = ofdef.OFPCML_NO_BUFFER
                                        )
                                    ]
                                )
                            ]
                        )
                    ]
            else:
                def _createicmpflows(ipaddress, macaddress, networkid):
                    return [
                        ofdef.ofp_flow_mod(
                            cookie = 0x2,
                            cookie_mask = 0xffffffffffffffff,
                            table_id = l3input,
                            command = ofdef.OFPFC_ADD,
                            # icmp to router matcher same as ip forward to router
                            # so priority + 1
                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                            buffer_id = ofdef.OFP_NO_BUFFER,
                            out_port = ofdef.OFPP_ANY,
                            out_group = ofdef.OFPG_ANY,
                            match = ofdef.ofp_match_oxm(
                                oxm_fields = [
                                    match_network(networkid),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST,mac_addr_bytes(macaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,ofdef.ETHERTYPE_IP),
                                    ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,ip4_addr_bytes(ipaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO,ofdef.IPPROTO_ICMP),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_TYPE,8),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_CODE,0)
                                ]
                            ),
                            instructions = [
                                ofdef.ofp_instruction_actions(
                                    actions = [
                                        ofdef.nx_action_reg_move(
                                            n_bits = 48,
                                            src = ofdef.OXM_OF_ETH_SRC,
                                            dst = ofdef.OXM_OF_ETH_DST
                                        ),
                                        ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(
                                                ofdef.OXM_OF_ETH_SRC,
                                                ofdef.mac_addr(macaddress)
                                            )
                                        ),
                                        ofdef.nx_action_reg_move(
                                            n_bits = 32,
                                            src = ofdef.OXM_OF_IPV4_SRC,
                                            dst = ofdef.OXM_OF_IPV4_DST
                                        ),
                                        ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(
                                                ofdef.OXM_OF_IPV4_SRC,
                                                ofdef.ip4_addr(ipaddress)
                                            )
                                        ),
                                        ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(
                                                ofdef.OXM_OF_ICMPV4_TYPE,
                                                ICMP_ECHOREPLY
                                            )
                                        ),
                                        ofdef.ofp_action_nw_ttl(
                                            nw_ttl = 128
                                        ),
                                        ofdef.ofp_action_output(
                                            port = ofdef.OFPP_IN_PORT
                                        )
                                    ]
                                )
                            ]
                        )
                    ]

            for subnet in lastsubnetsinfo.keys():
                if subnet not in currentsubnetsinfo\
                        or (subnet in currentsubnetsinfo and lastsubnetsinfo[subnet] != currentsubnetsinfo[subnet]):
                    # subnet remove  or subnet info changed , remove flow info
                    ip_address, mac_address, networkid, nid = lastsubnetsinfo[subnet]

                    remove_arp = {(ip_address,mac_address,networkid,True),}
                    await call_api(self, 'arpresponder', 'removeproxyarp', {'connection':connection,
                                                                              'arpentries': remove_arp})

                    cmds.extend(_deleteicmpflows(ip_address,mac_address,nid))

            await self.execute_commands(connection, cmds)

            for subnet in currentsubnetsinfo.keys():
                if subnet not in lastsubnetsinfo\
                        or (subnet in lastsubnetsinfo and lastsubnetsinfo[subnet] != currentsubnetsinfo[subnet]):

                    ip_address, mac_address, networkid, nid = currentsubnetsinfo[subnet]

                    add_arp = {(ip_address,mac_address,networkid,True),}
                    await call_api(self, 'arpresponder', 'createproxyarp', {'connection': connection,
                                                                               'arpentries': add_arp})

                    cmds.extend(_createicmpflows(ip_address,mac_address,nid))

            await self.execute_commands(connection, cmds)

        except Exception:
            self._logger.warning("Unexpected exception in icmp_flow_updater, ignore it! Continue",exc_info=True)

@defaultconfig
@depend(ofpportmanager.OpenflowPortManager,objectdb.ObjectDB)
class ICMPResponder(FlowBase):
    """
    Respond ICMP echo (ping) requests to the gateway
    """
    _tablerequest = (
        ("l3input",("l2input",),""),
        ("l2output",("l3input",),"")
    )
    # True : reply icmp ping with flow
    # False: reply icmp ping with controller PACKET_IN/PACKET_OUT
    #
    # Must use prepush=True with OpenvSwitch 2.5+
    #
    _default_prepush = False

    # "Gateway" responds with this MAC address
    _default_inroutermac = '1a:23:67:59:63:33'

    def __init__(self,server):
        super(ICMPResponder,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self._flowupdater = dict()

    async def _main(self):

        flowinit = FlowInitialize.createMatcher(_ismatch=lambda x: self.vhostbind is None or
                                                x.vhost in self.vhostbind)
        conndown = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                _ismatch=lambda x:self.vhostbind is None or
                                                x.createby.vhost in self.vhostbind)
        while True:
            ev, m = await M_(flowinit,conndown)
            if m is flowinit:
                c = ev.connection
                self.app_routine.subroutine(self._init_conn(c))
            if m is conndown:
                c = ev.connection
                self.app_routine.subroutine(self._remove_conn(c))

    async def _init_conn(self,conn):

        if conn in self._flowupdater:
            updater = self._flowupdater.pop(conn)
            updater.close()

        updater = ICMPResponderUpdater(conn,self)
        self._flowupdater[conn] = updater
        updater.start()

    async def _remove_conn(self,conn):

        if conn in self._flowupdater:
            updater = self._flowupdater.pop(conn)
            updater.close()



'''
Created on 2015/11/9

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, call_api
from vlcp.event.runnable import RoutineContainer
from vlcp.event import Event, withIndices
from vlcp.service.utils import knowledge
from uuid import uuid4
try:
    from Cookie import SimpleCookie, Morsel
except Exception:
    from http.cookies import SimpleCookie, Morsel
from vlcp.event.lock import Lock

@depend(knowledge.MemoryStorage)
@defaultconfig
class Session(Module):
    '''
    HTTP Session with cookies
    '''
    # Timeout for a session
    _default_timeout = 1800
    # Cookie name used in session id storage
    _default_cookiename = '_session_id'
    service = True
    class SessionObject(object):
        def __init__(self, sessionid):
            self.id = sessionid
            self.vars = {}
            self.lockseq = 0
            self.releaseseq = 0
    class SessionHandle(object):
        def __init__(self, sessionobj, container):
            self.sessionobj = sessionobj
            self.id = sessionobj.id
            self.vars = sessionobj.vars
            self._lock = Lock(sessionobj, container.scheduler)
            self.container = container
        async def lock(self):
            "Lock session"
            await self._lock.lock(self.container)
        def unlock(self):
            "Unlock session"
            self._lock.unlock()
    def __init__(self, server):
        '''
        Constructor
        '''
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.createAPI(api(self.start, self.apiroutine),
                       api(self.create, self.apiroutine),
                       api(self.get, self.apiroutine),
                       api(self.destroy, self.apiroutine))
    async def start(self, cookies, cookieopts = None):
        """
        Session start operation. First check among the cookies to find existed sessions;
        if there is not an existed session, create a new one.
        
        :param cookies: cookie header from the client
        
        :param cookieopts: extra options used when creating a new cookie
        
        :return: ``(session_handle, cookies)`` where session_handle is a SessionHandle object,
                 and cookies is a list of created Set-Cookie headers (may be empty)
        """
        c = SimpleCookie(cookies)
        sid = c.get(self.cookiename)
        create = True
        if sid is not None:
            sh = await self.get(sid.value)
            if sh is not None:
                return (self.SessionHandle(sh, self.apiroutine), [])
        if create:
            sh = await self.create()
            m = Morsel()
            m.key = self.cookiename
            m.value = sh.id
            m.coded_value = sh.id
            opts = {'path':'/', 'httponly':True}
            if cookieopts:
                opts.update(cookieopts)
                if not cookieopts['httponly']:
                    del cookieopts['httponly']
            m.update(opts)
            return (sh, [m])
    async def get(self, sessionid, refresh = True):
        """
        Get the seesion object of the session id
        
        :param sessionid: a session ID
        
        :param refresh: if True, refresh the expiration time of this session
        
        :return: Session object or None if not exists
        """
        return await call_api(self.apiroutine, 'memorystorage', 'get', {'key': __name__ + '.' + sessionid, 'timeout': self.timeout if refresh else None})
    async def create(self):
        """
        Create a new session object
        
        :return: Session handle for the created session object.
        """
        sid = uuid4().hex
        sobj = self.SessionObject(sid)
        await call_api(self.apiroutine, 'memorystorage', 'set', {'key': __name__ + '.' + sid, 'value': sobj, 'timeout': self.timeout})
        return self.SessionHandle(sobj, self.apiroutine)

    async def destroy(self, sessionid):
        """
        Destroy a session
        
        :param sessionid: session ID
        
        :return: a list of Set-Cookie headers to be sent to the client
        """
        await call_api(self.apiroutine, 'memorystorage', 'delete', {'key': __name__ + '.' + sessionid})
        m = Morsel()
        m.key = self.cookiename
        m.value = 'deleted'
        m.coded_value = 'deleted'
        opts = {'path':'/', 'httponly':True, 'max-age':0}
        m.update(opts)
        return [m]
    


'''
Created on 2018/7/2

:author: hubo
'''
class ModuleAPICallTimeoutException(Exception):
    pass


class APIRejectedException(Exception):
    pass


class StaleResultException(Exception):
    def __init__(self, result, desc = "Result is stale"):
        Exception.__init__(self, desc)
        self.result = result


class AsyncTransactionLockException(Exception):
    def __init__(self, info = None):
        Exception.__init__(self)
        self.info = info


class TransactionFailedException(Exception):
    pass


class TransactionRetryExceededException(TransactionFailedException):
    def __init__(self):
        TransactionFailedException.__init__(self, "Max retry exceeded")


class TransactionTimeoutException(TransactionFailedException):
    def __init__(self):
        TransactionFailedException.__init__(self, "Timeout exceeded")


class WalkKeyNotRetrieved(KeyError):
    pass



'''
Created on 2018/7/12

:author: hubo
'''
from vlcp.utils.exceptions import WalkKeyNotRetrieved


def ensure_keys(walk, *keys):
    """
    Use walk to try to retrieve all keys
    """
    all_retrieved = True
    for k in keys:
        try:
            walk(k)
        except WalkKeyNotRetrieved:
            all_retrieved = False
    return all_retrieved



"""
Physical network with type=local can have unlimited logical networks,
but it cannot have physical ports. So, logical ports in the same logical
network can access each other only when they are on the same host.
"""

from vlcp.server.module import Module,publicapi
from vlcp.event.runnable import RoutineContainer
from vlcp.config.config import defaultconfig
from vlcp.utils.networkplugin import createphysicalnetwork,\
    updatephysicalnetwork, default_physicalnetwork_keys, deletephysicalnetwork,\
    deletephysicalport, default_physicalport_keys, createlogicalnetwork,\
    default_logicalnetwork_keys, updatelogicalnetwork,\
    deletelogicalnetwork


@defaultconfig
class NetworkLocalDriver(Module):
    """
    Network driver for local networks. Local networks cannot have physical ports; logical networks
    in local networks do not have external connectivities, only endpoints on the same server can
    access each other.
    """
    def __init__(self,server):
        super(NetworkLocalDriver,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self.createAPI(
                       publicapi(self.createphysicalnetwork,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.updatephysicalnetwork,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.deletephysicalnetwork,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.createphysicalport,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.updatephysicalport,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.deletephysicalport,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.createlogicalnetwork,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.updatelogicalnetwork,
                                    criteria=lambda type: type == 'local'),
                       publicapi(self.deletelogicalnetwork,
                                    criteria=lambda type: type == "local"),
                       #used in IOprocessing module
                       publicapi(self.createioflowparts,
                                    criteria=lambda connection,logicalnetwork,
                                    physicalport,logicalnetworkid,physicalportid:
                                    logicalnetwork.physicalnetwork.type == "local")

                       )

    async def _main(self):
        self._logger.info("network_local_driver running ---")

    def createphysicalnetwork(self, type):
        # create an new physical network
        return createphysicalnetwork(type), default_physicalnetwork_keys
    
    def updatephysicalnetwork(self, type):
        # update a physical network
        return updatephysicalnetwork(), None
    
    def deletephysicalnetwork(self, type):
        return deletephysicalnetwork(), default_physicalnetwork_keys
    
    def createphysicalport(self, type):
        raise ValueError("A physical network with type=local cannot have physical ports")
        
    def updatephysicalport(self, type):
        raise ValueError("A physical network with type=local cannot have physical ports")
   
    def deletephysicalport(self, type):
        return deletephysicalport(), default_physicalport_keys

    def createlogicalnetwork(self,type):
        return createlogicalnetwork(), default_logicalnetwork_keys

    def updatelogicalnetwork(self,type):
        return updatelogicalnetwork(), None

    def deletelogicalnetwork(self,type):
        return deletelogicalnetwork(), default_logicalnetwork_keys

    def createioflowparts(self,connection,logicalnetwork,physicalport,logicalnetworkid,physicalportid):
        return [], [], [], [], []




'''
Created on 2015/06/01

:author: hubo
'''
from __future__ import print_function, absolute_import, division 
from collections import OrderedDict

class MatchTree(object):
    '''
    A dictionary tree for fast event match
    '''
    def __init__(self, parent = None):
        '''
        Constructor
        '''
        self.index = {}
        #self.matchers = OrderedDict()
        # JIT friendly
        self.matchers_list = []
        self.matchers_dict = None
        self._use_dict = False
        self.parent = parent
        if parent is not None:
            self.depth = parent.depth + 1
        else:
            self.depth = 0
    
    def subtree(self, matcher, create = False):
        '''
        Find a subtree from a matcher
        
        :param matcher: the matcher to locate the subtree. If None, return the root of the tree.
        
        :param create: if True, the subtree is created if not exists; otherwise return None if not exists
        '''
        if matcher is None:
            return self
        current = self
        for i in range(self.depth, len(matcher.indices)):
            ind = matcher.indices[i]
            if ind is None:
                # match Any
                if hasattr(current, 'any'):
                    current = current.any
                else:
                    if create:
                        cany = MatchTree(current)
                        cany.parentIndex = None
                        current.any = cany
                        current = cany
                    else:
                        return None
            else:
                current2 = current.index.get(ind)
                if current2 is None:
                    if create:
                        cind = MatchTree(current)
                        cind.parentIndex = ind 
                        current.index[ind] = cind
                        current = cind
                    else:
                        return None
                else:
                    current = current2
        return current        
    def insert(self, matcher, obj):
        '''
        Insert a new matcher
        
        :param matcher: an EventMatcher
        
        :param obj: object to return
        '''
        current = self.subtree(matcher, True)
        #current.matchers[(obj, matcher)] = None
        if current._use_dict:
            current.matchers_dict[(obj, matcher)] = None
        else:
            current.matchers_list.append((obj, matcher))
        return current
    def remove(self, matcher, obj):
        '''
        Remove the matcher
        
        :param matcher: an EventMatcher
        
        :param obj: the object to remove
        '''
        current = self.subtree(matcher, False)
        if current is None:
            return
        # Assume that this pair only appears once
        if current._use_dict:
            try:
                del current.matchers_dict[(obj, matcher)]
            except KeyError:
                pass
        else:
            if len(current.matchers_list) > 10:
                # Convert to dict
                current.matchers_dict = OrderedDict((v, None) for v in current.matchers_list
                                                    if v != (obj, matcher))
                current.matchers_list = None
                current._use_dict = True
            else:
                try:
                    current.matchers_list.remove((obj, matcher))
                except ValueError:
                    pass
        while ((not current.matchers_dict) if current._use_dict
               else (not current.matchers_list))\
                and not current.matchers_dict\
                and not hasattr(current,'any')\
                and not current.index and current.parent is not None:
            # remove self from parents
            ind = current.parentIndex
            if ind is None:
                del current.parent.any
            else:
                del current.parent.index[ind]
            p = current.parent
            current.parent = None
            current = p
    def matchesWithMatchers(self, event):
        '''
        Return all matches for this event. The first matcher is also returned for each matched object.
        
        :param event: an input event
        '''
        ret = []
        self._matches(event, set(), ret)
        return tuple(ret)
    def matches(self, event):
        '''
        Return all matches for this event. The first matcher is also returned for each matched object.
        
        :param event: an input event
        '''
        ret = []
        self._matches(event, set(), ret)
        return tuple(r[0] for r in ret)
    def _matches(self, event, duptest, retlist):
        # 1. matches(self.index[ind], event)
        # 2. matches(self.any, event)
        # 3. self.matches
        if self.depth < len(event.indices):
            ind = event.indices[self.depth]
            if ind in self.index:
                self.index[ind]._matches(event, duptest, retlist)
            if hasattr(self, 'any'):
                self.any._matches(event, duptest, retlist)
        if self._use_dict:
            for v in self.matchers_dict:
                o, m = v
                if o not in duptest and m.judge(event):
                    duptest.add(o)
                    retlist.append(v)
        else:
            for v in self.matchers_list:
                o, m = v
                if o not in duptest and m.judge(event):
                    duptest.add(o)
                    retlist.append(v)
            
    def matchfirst(self, event):
        '''
        Return first match for this event
        
        :param event: an input event
        '''
        # 1. matches(self.index[ind], event)
        # 2. matches(self.any, event)
        # 3. self.matches
        if self.depth < len(event.indices):
            ind = event.indices[self.depth]
            if ind in self.index:
                m = self.index[ind].matchfirst(event)
                if m is not None:
                    return m
            if hasattr(self, 'any'):
                m = self.any.matchfirst(event)
                if m is not None:
                    return m
        if self._use_dict:
            for o, m in self.matchers_dict:
                if m is None or m.judge(event):
                    return o
        else:
            for o, m in self.matchers_list:
                if m is None or m.judge(event):
                    return o
                
    def matchfirstwithmatcher(self, event):
        '''
        Return first match with matcher for this event
        
        :param event: an input event
        '''
        # 1. matches(self.index[ind], event)
        # 2. matches(self.any, event)
        # 3. self.matches
        if self.depth < len(event.indices):
            ind = event.indices[self.depth]
            if ind in self.index:
                m = self.index[ind].matchfirstwithmatcher(event)
                if m is not None:
                    return m
            if hasattr(self, 'any'):
                m = self.any.matchfirstwithmatcher(event)
                if m is not None:
                    return m
        if self._use_dict:
            for v in self.matchers_dict:
                o, m = v
                if m is None or m.judge(event):
                    return v
        else:
            for v in self.matchers_list:
                o, m = v
                if m is None or m.judge(event):
                    return v                
        return None
    

class EventTree(object):
    '''
    Store events; match matchers
    '''
    def __init__(self, parent = None, branch = 5):
        '''
        Constructor
        '''
        self.events = []
        self.subevents = []
        self.parent = parent
        if parent is not None:
            self.depth = parent.depth + 1
        else:
            self.depth = 0
        self.branch = 5
    def subtree(self, event, create = False):
        '''
        Find a subtree from an event
        '''
        current = self
        for i in range(self.depth, len(event.indices)):
            if not hasattr(current, 'index'):
                return current
            ind = event.indices[i]
            if create:
                current = current.index.setdefault(ind, EventTree(current, self.branch))
                current.parentIndex = ind
            else:
                current = current.index.get(ind)
                if current is None:
                    return None
        return current
    def insert(self, event):
        current = self.subtree(event, True)
        if current.depth == len(event.indices):
            current.events.append(event)
        else:
            current.subevents.append(event)
            if len(current.subevents) > self.branch:
                current.index = {}
                for e in current.subevents:
                    current.insert(e)
                del current.subevents[:]
    def _returnAll(self, ret):
        ret.extend(self.events)
        if hasattr(self, 'index'):
            for st in self.index.values():
                st._returnAll(ret)
        else:
            ret.extend(self.subevents)
    def _removeFromParent(self):
        current = self
        while current.parent is not None:
            del current.parent.index[current.parentIndex]
            p = current.parent
            current.parent = None
            current = p
            if current.index or current.events:
                break
        if hasattr(current, 'index') and not current.index:
            del current.index
    def _findAndRemove(self, matcher, ret):
        current = self
        while current.depth < len(matcher.indices):
            ind = matcher.indices[current.depth]
            if not hasattr(current, 'index'):
                newsub = []
                for e in current.subevents:
                    if matcher.isMatch(e, current.depth):
                        ret.append(e)
                    else:
                        newsub.append(e)
                current.subevents = newsub
                if not current.subevents and not current.events:
                    current._removeFromParent()
                return
            elif ind is None:
                for st in current.index.values():
                    st._findAndRemove(matcher, ret)
                return
            else:
                if ind not in current.index:
                    return
                current = current.index[ind]
        current._returnAll(ret)
        current._removeFromParent()
    def findAndRemove(self, matcher):
        ret = []
        self._findAndRemove(matcher, ret)
        return tuple(ret)
    def remove(self, event):
        current = self.subtree(event)
        if current.depth == len(event.indices):
            current.events.remove(event)
        else:
            current.subevents.remove(event)

            



'''
Created on 2016/3/21

:author: hubo
'''

from vlcp.server.module import Module, call_api, depend, ModuleNotification,\
    ModuleLoadStateChanged, api, proxy
from vlcp.config import defaultconfig
import vlcp.service.connection.redisdb as redisdb
from vlcp.event.event import withIndices, Event, M_
from vlcp.event.runnable import RoutineContainer
from vlcp.event.connection import ConnectionResetException
from vlcp.event.core import syscall_removequeue, QuitException
import json
from time import time
from zlib import compress, decompress, error as zlib_error
import functools
import uuid
import logging
from contextlib import closing


@withIndices('notifier', 'transactid', 'keys', 'reason', 'fromself')
class UpdateNotification(Event):
    UPDATED = 'updated'
    RESTORED = 'restored'


@withIndices('notifier', 'stage')
class ModifyListen(Event):
    SUBSCRIBE = 'subscribe'
    LISTEN = 'listen'


def _bytes(s):
    if isinstance(s, bytes):
        return s
    else:
        return s.encode('utf-8')


def _str(s):
    if isinstance(s, str):
        return s
    else:
        return s.decode('utf-8')


class _Notifier(RoutineContainer):
    _logger = logging.getLogger(__name__ + '.Notifier')

    def __init__(self, vhostbind, prefix, scheduler=None, singlecastlimit = 256, deflate = False):
        RoutineContainer.__init__(self, scheduler=scheduler, daemon=False)
        self.vhostbind = vhostbind
        self.prefix = _bytes(prefix)
        self._matchers = {}
        self._publishkey = uuid.uuid1().hex
        self._publishno = 1
        self._publish_wait = set()
        self._matchadd_wait = set()
        self._matchremove_wait = set()
        self._singlecastlimit = singlecastlimit
        self._deflate = deflate

    async def main(self):
        try:
            timestamp = '%012x' % (int(time() * 1000),) + '-'
            transactno = 1
            client, encoder, decoder = await call_api(self, 'redisdb', 'getclient', {'vhost':self.vhostbind})
            try:
                s_matchers = await client.subscribe(self, self.prefix)
            except Exception:
                _no_connection_start = True
            else:
                _no_connection_start = False
            self._matchers[b''] = s_matchers[0]
            if self._deflate:
                oldencoder = encoder
                olddecoder = decoder
                def encoder(x):
                    return compress(oldencoder(x))
                def decoder(x):
                    try:
                        return olddecoder(decompress(x))
                    except zlib_error:
                        return olddecoder(x)
            if not _no_connection_start:
                self.subroutine(self._modifier(client), True, "modifierroutine")
            listen_modify = ModifyListen.createMatcher(self, ModifyListen.LISTEN)
            connection_down = client.subscribe_state_matcher(self, False)
            connection_up = client.subscribe_state_matcher(self, True)
            module_loaded = ModuleLoadStateChanged.createMatcher(state = ModuleLoadStateChanged.LOADED,
                                                 _ismatch = lambda x: x._instance.getServiceName() == 'redisdb')
            matchers = tuple(self._matchers.values()) + (listen_modify, connection_down)
            last_transact = None
            while True:
                if not _no_connection_start:
                    ev, m = await M_(*matchers)
                if not _no_connection_start and m is listen_modify:
                    matchers = tuple(self._matchers.values()) + (listen_modify, connection_down)
                elif _no_connection_start or m is connection_down:
                    # Connection is down, wait for restore
                    # The module may be reloaded
                    if _no_connection_start:
                        recreate_matchers = True
                    else:
                        recreate_matchers = False
                    last_transact = None
                    while True:
                        _, m = await M_(connection_up, module_loaded)
                        if m is module_loaded:
                            self.terminate(self.modifierroutine)
                            client, encoder, decoder = await call_api(self, 'redisdb', 'getclient', {'vhost':self.vhostbind})
                            if self._deflate:
                                oldencoder = encoder
                                olddecoder = decoder
                                def encoder(x):
                                    return compress(oldencoder(x))
                                def decoder(x):
                                    try:
                                        return olddecoder(decompress(x))
                                    except zlib_error:
                                        return olddecoder(x)
                            # Recreate listeners
                            connection_down = client.subscribe_state_matcher(self, False)
                            connection_up = client.subscribe_state_matcher(self, True)
                            try:
                                s_matchers = await client.subscribe(self, *tuple(self._matchers.keys()))
                            except Exception:
                                recreate_matchers = True
                                continue
                            else:
                                self._matchers = dict(zip(self._matchers.keys(), s_matchers))
                                self.subroutine(self._modifier(client), True, "modifierroutine")
                                matchers = tuple(self._matchers.values()) + (listen_modify, connection_down)
                                break
                        else:
                            if recreate_matchers:
                                try:
                                    s_matchers = await client.subscribe(self, *[self.prefix + k for k in self._matchers.keys()])
                                except Exception:
                                    recreate_matchers = True
                                    continue
                                else:
                                    self._matchers = dict(zip(self._matchers.keys(), s_matchers))
                                    self.subroutine(self._modifier(client), True, "modifierroutine")
                                    matchers = tuple(self._matchers.values()) + (listen_modify, connection_down)
                                    break
                            else:
                                matchers = tuple(self._matchers.values()) + (listen_modify, connection_down)
                                break
                    if self._publish_wait:
                        self.subroutine(self.publish())
                    transactid = '%s%016x' % (timestamp, transactno)
                    transactno += 1
                    async def send_restore_notify(transactid):
                        if self._matchadd_wait or self._matchremove_wait:
                            # Wait for next subscribe success
                            await self.wait_with_timeout(1, ModifyListen.createMatcher(self, ModifyListen.LISTEN))
                        await self.wait_for_send(
                                UpdateNotification(
                                    self,
                                    transactid,
                                    tuple(self._matchers.keys()),
                                    UpdateNotification.RESTORED,
                                    False,
                                    extrainfo = None)
                                )
                    self.subroutine(send_restore_notify(transactid), False)
                else:
                    transact = decoder(ev.message)
                    if transact['id'] == last_transact:
                        # Ignore duplicated updates
                        continue
                    last_transact = transact['id']
                    pubkey, sep, pubno = last_transact.partition('-')
                    fromself = (sep and pubkey == self._publishkey)
                    transactid = '%s%016x' % (timestamp, transactno)
                    transactno += 1
                    self.subroutine(self.wait_for_send(
                                            UpdateNotification(
                                                self,
                                                transactid,
                                                tuple(_bytes(k) for k in transact['keys']),
                                                UpdateNotification.UPDATED,
                                                fromself,
                                                extrainfo = transact.get('extrainfo'))
                                        ),
                                    False)
        finally:
            if hasattr(self ,'modifierroutine') and self.modifierroutine:
                self.terminate(self.modifierroutine)

    async def _modifier(self, client):
        try:
            modify_matcher = ModifyListen.createMatcher(self, ModifyListen.SUBSCRIBE)
            while True:
                try:
                    while self._matchadd_wait or self._matchremove_wait:
                        if self._matchadd_wait:
                            # Subscribe new keys
                            current_add = set(self._matchadd_wait)
                            self._matchadd_wait.clear()
                            add_keys = list(current_add.difference(self._matchers.keys()))
                            try:
                                s_matchers = await client.subscribe(self, *[self.prefix + k for k in add_keys])
                            except:
                                # Return to matchadd
                                self._matchadd_wait.update(current_add.difference(self._matchremove_wait))
                                raise
                            else:
                                self._matchers.update(zip(add_keys, s_matchers))
                                await self.wait_for_send(ModifyListen(self, ModifyListen.LISTEN))
                        if self._matchremove_wait:
                            # Unsubscribe keys
                            current_remove = set(self._matchremove_wait)
                            self._matchremove_wait.clear()
                            del_keys = list(current_remove.intersection(self._matchers.keys()))
                            try:
                                await client.unsubscribe(self, *[self.prefix + k for k in del_keys])
                            except:
                                # Return to matchremove
                                self._matchremove_wait.update(current_remove.difference(self._matchadd_wait))
                                raise
                            else:
                                for k in del_keys:
                                    del self._matchers[k]
                                await self.wait_for_send(ModifyListen(self, ModifyListen.LISTEN))
                    await modify_matcher
                except (IOError, ConnectionResetException):
                    # Wait for connection resume
                    connection_up = client.subscribe_state_matcher(self)
                    await connection_up
        finally:
            self.subroutine(self._clearup(client, list(self._matchers.keys())))

    async def _clearup(self, client, keys):
        try:
            if not self.scheduler.quitting:
                await client.unsubscribe(self, *[self.prefix + k for k in keys])
        except Exception:
            pass

    async def add_listen(self, *keys):
        keys = [_bytes(k) for k in keys]
        self._matchremove_wait.difference_update(keys)
        self._matchadd_wait.update(keys)
        await self.wait_for_send(ModifyListen(self, ModifyListen.SUBSCRIBE))

    async def remove_listen(self, *keys):        
        keys = [_bytes(k) for k in keys]
        self._matchadd_wait.difference_update(keys)
        self._matchremove_wait.update(keys)
        await self.wait_for_send(ModifyListen(self, ModifyListen.SUBSCRIBE))

    async def publish(self, keys = (), extrainfo = None):
        keys = [_bytes(k) for k in keys]
        if self._publish_wait:
            merged_keys = list(self._publish_wait.union(keys))
            self._publish_wait.clear()
        else:
            merged_keys = list(keys)
        if not merged_keys:
            return
        client, encoder, _ = await call_api(self, 'redisdb', 'getclient', {'vhost':self.vhostbind})
        transactid = '%s-%016x' % (self._publishkey, self._publishno)
        self._publishno += 1
        msg = encoder({'id':transactid, 'keys':[_str(k) for k in merged_keys], 'extrainfo': extrainfo})
        try:
            if len(merged_keys) > self._singlecastlimit:
                await client.execute_command(self, 'PUBLISH', self.prefix, msg)
            else:
                await client.batch_execute(self, *((('MULTI',),) +
                                                    tuple(('PUBLISH', self.prefix + k, msg) for k in merged_keys) +
                                                    (('EXEC',),)))
        except (IOError, ConnectionResetException):
            self._logger.warning('Following keys are not published because exception occurred, delay to next publish: %r', merged_keys, exc_info = True)
            self._publish_wait.update(merged_keys)

    def notification_matcher(self, fromself = None):
        if fromself is None:
            return UpdateNotification.createMatcher(self)
        else:
            return UpdateNotification.createMatcher(notifier = self, fromself = fromself)


@defaultconfig
@depend(redisdb.RedisDB)
class RedisNotifier(Module):
    """
    Update notification with Redis Pub/Sub
    """
    # Bind to RedisDB vHost
    _default_vhostbind = ''
    # Use this prefix for subscribe channels
    _default_prefix = 'vlcp.updatenotifier.'
    # If a notification contains more than .singlecastlimit keys, do not replicate the notification
    # to all channels; only broadcast it once in the public channel.
    _default_singlecastlimit = 256
    # Use an extra deflate compression on notification, should not be necessary if you already have
    # module.redisdb.deflate=True
    _default_deflate = False
    def __init__(self, server):
        Module.__init__(self, server)
        self.create_api(api(self.createnotifier))

    async def load(self, container):
        self.scheduler.queue.addSubQueue(999, ModifyListen.createMatcher(), "redisnotifier_modifylisten")
        await Module.load(self, container)

    async def unload(self, container, force=False):
        await Module.unload(self, container, force=force)
        await container.syscall_noreturn(syscall_removequeue(self.scheduler.queue, "redisnotifier_modifylisten"))

    def createnotifier(self):
        "Create a new notifier object"
        n = _Notifier(self.vhostbind, self.prefix, self.scheduler, self.singlecastlimit, self.deflate)
        n.start()
        return n


UpdateNotifier = proxy('UpdateNotifier', RedisNotifier)



'''
Created on 2016/12/1

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, ModuleNotification, call_api,\
    send_api
from vlcp.event.runnable import RoutineContainer
import vlcp.utils.ovsdb as ovsdb
import vlcp.service.connection.jsonrpcserver as jsonrpcserver
from vlcp.protocol.jsonrpc import JsonRPCErrorResultException,\
    JsonRPCConnectionStateEvent, JsonRPCNotificationEvent
from vlcp.event.event import Event, withIndices, M_
from contextlib import closing
from vlcp.event.connection import ConnectionResetException
from vlcp.utils.vxlandiscover import lognet_vxlan_walker, update_vxlaninfo,\
    get_broadcast_ips
import vlcp.service.kvdb.objectdb as objectdb
from vlcp.utils.networkmodel import LogicalNetwork, VXLANEndpointSet,\
    LogicalNetworkMap, LogicalPortVXLANInfo, LogicalPort
from vlcp.utils.dataobject import multiwaitif
from vlcp.utils.ethernet import ip4_addr

@withIndices('connection')
class VtepConnectionSynchronized(Event):
    pass


@withIndices('state', 'name', 'connection', 'connmark', 'vhost')
class VtepPhysicalSwitchStateChanged(Event):
    UP = 'up'
    DOWN = 'down'


@withIndices('identifier')
class _DataUpdateEvent(Event):
    pass

def _check_transact_result(result, operations):
    if any('error' in r for r in result if r is not None):
        err_info = next((r['error'],
                     operations[i] if i < len(operations) else None)
                    for i,r in enumerate(result)
                    if r is not None and 'error' in r)
        raise JsonRPCErrorResultException('Error in OVSDB select operation: %r. Corresponding operation: %r' % err_info)

@defaultconfig
@depend(jsonrpcserver.JsonRPCServer, objectdb.ObjectDB)
class VtepController(Module):
    """
    Controll a physical switch which supports OVSDB hardware_vtep protocol.
    """
    # Default bind controller to OVSDB vHost
    _default_vhostbind = ['vtep']
    # Recycle unused logical switches from hardware_vtep OVSDB
    _default_recycleinterval = 300
    # Every logical network has a broadcast list which contains all related server nodes,
    # this is the refresh interval to keep current node in the list
    _default_refreshinterval = 3600
    # When a logical port is migrating from one node to another, there may be multiple
    # instances of the logical port on different or same nodes. This is the maximum time
    # allowed for the migration.
    _default_allowedmigrationtime = 120
    # When there are multiple controllers started for the same hardware_vtep OVSDB,
    # every instance will try to acquire this lock in the OVSDB, and the one who
    # acquired the lock will be the master. 
    _default_masterlock = "vlcp_vtepcontroller_masterlock"
    # Prepush unicast information for MACs on other nodes. If this is not set, the
    # physical switch should support and enable MAC-learning on VXLAN.
    _default_prepush = True
    def __init__(self, server):
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self.createAPI(api(self.listphysicalports, self.apiroutine),
                       api(self.listphysicalswitches, self.apiroutine),
                       api(self.updatelogicalswitch, self.apiroutine),
                       api(self.unbindlogicalswitch, self.apiroutine),
                       api(self.unbindphysicalport, self.apiroutine))
        self._physical_switchs = {}
        self._connection_ps = {}
        self._synchronized = False
        self._monitor_routines = set()
        self._requestid = 0
        
    async def _wait_for_sync(self):
        if not self._synchronized:
            await ModuleNotification.createMatcher(self.getServiceName(), 'synchronized')
    
    async def _monitor_conn(self, conn, notification = False):
        current_routine = self.apiroutine.currentroutine
        self._monitor_routines.add(current_routine)
        ls_monitor = self.apiroutine.subroutine(self._monitor_logicalswitches(conn))
        initialized = False
        protocol = conn.protocol
        def _process_result(result):
            if 'old' in result:
                # Removed or modified
                if 'name' in result['old']:
                    oldname = result['old']['name']
                else:
                    oldname = result['new']['name']
                try:
                    del self._physical_switchs[oldname]
                except KeyError:
                    pass
                try:
                    self._connection_ps[conn].discard(oldname)
                except KeyError:
                    pass
                if 'new' not in result or 'name' in result['old']:
                    self.scheduler.emergesend(VtepPhysicalSwitchStateChanged(VtepPhysicalSwitchStateChanged.DOWN,
                                                                             oldname,
                                                                             conn,
                                                                             conn.connmark,
                                                                             protocol.vhost
                                                                             ))
            if 'new' in result:
                # Created or modified
                nr = result['new']
                self._connection_ps.setdefault(conn, set()).add(nr['name'])
                tunnel_ips = ovsdb.getlist(nr['tunnel_ips'])
                if not tunnel_ips:
                    self._logger.warning('Physical switch %r does not have a configured tunnel IP', )
                    tunnelip = None
                else:
                    tunnelip = tunnel_ips[0]
                self._physical_switchs[nr['name']] = (conn, tunnelip)
                if 'old' not in result or 'name' in result['old']:
                    self.scheduler.emergesend(VtepPhysicalSwitchStateChanged(VtepPhysicalSwitchStateChanged.UP,
                                                                             nr['name'],
                                                                             conn,
                                                                             conn.connmark,
                                                                             protocol.vhost
                                                                             ))                
        try:
            method, params = ovsdb.monitor(
                                'hardware_vtep',
                                'vlcp_vtepcontroller_physicalswitch_monitor',
                                {
                                    'Physical_Switch':
                                        ovsdb.monitor_request(['name', 'tunnel_ips'], True, True, True, True)
                                })
            try:
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
            except JsonRPCErrorResultException:
                # The monitor is already set, cancel it first
                method2, params2 = ovsdb.monitor_cancel('vlcp_vtepcontroller_physicalswitch_monitor')
                await protocol.querywithreply(method2, params2, conn, self.apiroutine, False)
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
            for r in result['Physical_Switch'].values():
                _process_result(r)
            initialized = True
            if notification:
                self.scheduler.emergesend(VtepConnectionSynchronized(conn))
            monitor_matcher = ovsdb.monitor_matcher(conn, 'vlcp_vtepcontroller_physicalswitch_monitor')
            conn_state = protocol.statematcher(conn)
            while True:
                ev, m = await M_(monitor_matcher, conn_state)
                if m is conn_state:
                    break
                else:
                    for r in ev.params[1]['Physical_Switch'].values():
                        _process_result(r)
        except Exception:
            self._logger.warning('Initialize OVSDB vtep connection failed, maybe this is not a valid vtep endpoint',
                                 exc_info = True)
        finally:
            ls_monitor.close()
            if notification and not initialized:
                self.scheduler.emergesend(VtepConnectionSynchronized(conn))
            if conn in self._connection_ps:
                ps_set = self._connection_ps[conn]
                for ps in ps_set:
                    if ps in self._physical_switchs:
                        del self._physical_switchs[ps]
                del self._connection_ps[conn]
            self._monitor_routines.discard(current_routine)

    async def _monitor_logicalswitches(self, conn):
        protocol = conn.protocol
        lockid = self.masterlock
        # There may be multiple vtep controllers running, only the one which successfully acquired the
        # master lock should update the ObjectDB
        method, params = ovsdb.lock(lockid)
        async def _unlock():
            method, params = ovsdb.unlock(lockid)
            if conn.connected:
                try:
                    await protocol.querywithreply(method, params, conn, self.apiroutine)
                except Exception:
                    # Any result is acceptable, including: error result; connection down; etc.
                    # We don't report an exception that we cannot do anything for it.
                    pass
        try:
            result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
        except JsonRPCErrorResultException:
            # For some reason the lock is not released? let's try it out
            await _unlock()
            result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
        switch_routines = {}
        try:
            if not result['locked']:
                # Wait for lock
                lock_notification = JsonRPCNotificationEvent.createMatcher(
                                            "locked",
                                            conn,
                                            conn.connmark,
                                            _ismatch = lambda x: x.params[0] == lockid)
                conn_matcher = protocol.statematcher(conn)
                _, m = await M_(lock_notification, conn_matcher)
                if m is conn_matcher:
                    return
            # Now we acquired the lock, start monitoring the logical switches
            method, params = ovsdb.monitor(
                                'hardware_vtep',
                                'vlcp_vtepcontroller_logicalswitch_monitor',
                                {
                                    'Logical_Switch':
                                        ovsdb.monitor_request(['name'], True, True, True, False)
                                })
            try:
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
            except JsonRPCErrorResultException:
                # The monitor is already set, cancel it first
                method2, params2 = ovsdb.monitor_cancel('vlcp_vtepcontroller_logicalswitch_monitor')
                await protocol.querywithreply(method2, params2, conn, self.apiroutine, False)
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
            if 'Logical_Switch' in result:
                for k, r in result['Logical_Switch'].items():
                    switch_routines[k] = self.apiroutine.subroutine(self._update_logicalswitch_info(conn, r['new']['name'], k))
            monitor_matcher = ovsdb.monitor_matcher(conn, 'vlcp_vtepcontroller_logicalswitch_monitor')
            conn_state = protocol.statematcher(conn)
            while True:
                ev, m = await M_(monitor_matcher, conn_state)
                if m is conn_state:
                    break
                updates = ev.params[1]
                if 'Logical_Switch' in updates:
                    for k,r in ev.params[1]['Logical_Switch'].items():
                        if 'old' in r:
                            if k in switch_routines:
                                switch_routines[k].close()
                                del switch_routines[k]
                        else:
                            if k in switch_routines:
                                switch_routines[k].close()
                            switch_routines[k] = self.apiroutine.subroutine(self._update_logicalswitch_info(conn, r['new']['name'], k))
        finally:
            for r in switch_routines.values():
                r.close()
            self.apiroutine.subroutine(_unlock(), False)
    
    async def _push_logicalswitch_endpoints(self, conn, logicalswitchid):
        while True:
            if conn in self._connection_ps:
                ps_list = [(name, self._physical_switchs[name][1])
                           for name in self._connection_ps[conn]]
                for name, tunnel_ip in ps_list:
                    await update_vxlaninfo(
                                    self.apiroutine,
                                    {logicalswitchid: tunnel_ip},
                                    {},
                                    {},
                                    conn.protocol.vhost,
                                    name,
                                    'hardware_vtep',
                                    self.allowedmigrationtime,
                                    self.refreshinterval)
            await self.apiroutine.wait_with_timeout(self.refreshinterval)
    
    async def _poll_logicalswitch_endpoints(self, conn, logicalswitchid, lsuuid):
        protocol = conn.protocol
        walker = lognet_vxlan_walker(self.prepush)
        self._requestid += 1
        requestid = ('vtepcontroller', self._requestid)
        lognet_key = LogicalNetwork.default_key(logicalswitchid)
        endpointset_key = VXLANEndpointSet.default_key(logicalswitchid)
        lognetmap_key = LogicalNetworkMap.default_key(logicalswitchid)
        rewalk_keys = (lognet_key, endpointset_key, lognetmap_key)
        _update_set = set()
        _detect_update_routine = None
        identifier = object()
        async def _detect_update(saved_result):
            while True:
                updated_values, _ = await multiwaitif(saved_result, self.apiroutine, lambda x,y: True, True)
                if not _update_set:
                    self.apiroutine.scheduler.emergesend(_DataUpdateEvent(identifier))
                _update_set.update(updated_values)
        update_matcher = _DataUpdateEvent.createMatcher(identifier)
        _savedkeys = ()
        _savedresult = ()
        # These are: set of network tunnelips; dictionary for mac: tunnelip; last_physical_switch
        _last_endpoints = [set(), {}, None]
        
        async def _update_hardware_vtep(saved_result):
            if conn not in self._connection_ps or not self._connection_ps[conn]:
                _last_endpoints[:] = [set(), {}, None]
                return
            # There should be only one physical switch in this connection
            ps = next(iter(self._connection_ps[conn]))
            allobjs = [v for v in saved_result if v is not None and not v.isdeleted()]
            lognet_list = [o for o in allobjs if o.getkey() == lognet_key]
            if not lognet_list or not hasattr(lognet_list[0], 'vni'):
                # Logical network maybe deleted, or not a VXLAN network, do not need to update
                return
            vni = lognet_list[0].vni
            endpointset_list = [v for v in allobjs if v.getkey() == endpointset_key]
            local_ip = self._physical_switchs[ps][1]
            if endpointset_list:
                broadcast_ips = set(ip for ip,_ in get_broadcast_ips(endpointset_list[0], local_ip,
                                                  protocol.vhost,
                                                  ps,
                                                  'hardware_vtep'))
            else:
                broadcast_ips = set()
            if self.prepush:
                logport_tunnel_ip_dict = dict((o.id, o.endpoints[0]['tunnel_dst'])
                                              for o in allobjs
                                              if o.isinstance(LogicalPortVXLANInfo) \
                                              and hasattr(o, 'endpoints') and o.endpoints \
                                              and (o.endpoints[0]['vhost'],
                                                   o.endpoints[0]['systemid'],
                                                   o.endpoints[0]['bridge'])
                                                != (protocol.vhost, ps, 'hardware_vtep'))
                mac_ip_dict = dict((p.mac_address, logport_tunnel_ip_dict[p.id])
                                   for p in allobjs
                                   if p.isinstance(LogicalPort) \
                                   and hasattr(p, 'mac_address') \
                                   and p.id in logport_tunnel_ip_dict)
            else:
                mac_ip_dict = {}
            if _last_endpoints[2] == ps:
                # Update
                update = True
                broadcasts_updated = (broadcast_ips != _last_endpoints[0])
                add_ucasts = dict((m, mac_ip_dict[m])
                                  for m in mac_ip_dict
                                  if m not in _last_endpoints[1] or \
                                    _last_endpoints[1][m] != mac_ip_dict[m])
                remove_ucasts = dict((m, _last_endpoints[1][m])
                                     for m in _last_endpoints[1]
                                     if m not in mac_ip_dict)
            else:
                update = False
                broadcasts_updated = True
                remove_broadcasts = set()
                add_ucasts = mac_ip_dict
                remove_ucasts = {}
            using_ips = list(broadcast_ips.union(add_ucasts.values()))
            try:
                while True:
                    # Check and set
                    check_operates = [ovsdb.select('Logical_Switch',
                                                     [["_uuid", "==", ovsdb.uuid(lsuuid)],
                                                      ["name", "==", logicalswitchid]],
                                                     ["_uuid", "name", "tunnel_key"]),
                                      ovsdb.select('Mcast_Macs_Remote',
                                                     [["MAC", "==", "unknown-dst"],
                                                      ["logical_switch", "==", ovsdb.uuid(lsuuid)]],
                                                     ["_uuid", "locator_set"])]
                    # Check if each destination location exists
                    check_operates.extend(ovsdb.select('Physical_Locator',
                                                       [["dst_ip", "==", ip],
                                                        ["encapsulation_type", "==", "vxlan_over_ipv4"]],
                                                       ["_uuid",  "dst_ip"])
                                          for ip in using_ips)
                    method, params = ovsdb.transact('hardware_vtep',
                                                    *check_operates)
                    result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                    _check_transact_result(result, check_operates)
                    if not result[0]['rows']:
                        # Logical switch is removed
                        break
                    wait_operates = [ovsdb.wait('Logical_Switch',
                                               [["_uuid", "==", ovsdb.uuid(lsuuid)]],
                                               ["name"],
                                               [{"name": logicalswitchid}], True, 0)]
                    set_operates = []
                    # Check destinations
                    locator_uuid_dict = {}
                    if result[0]['rows'][0]['tunnel_key'] != vni:
                        set_operates.append(ovsdb.update('Logical_Switch',
                                               [["_uuid", "==", ovsdb.uuid(lsuuid)]],
                                               {"tunnel_key": vni}))
                    for r,ip in zip(result[2:len(check_operates)], using_ips):
                        if r['rows']:
                            locator = r['rows'][0]
                            # This locator is already created
                            locator_uuid_dict[ip] = locator['_uuid']
                            wait_operates.append(ovsdb.wait('Physical_Locator',
                                                       [["_uuid", "==", locator['_uuid']]],
                                                       ["dst_ip", "encapsulation_type"],
                                                       [{"dst_ip": ip,
                                                         "encapsulation_type": "vxlan_over_ipv4"}],
                                                        True, 0))
                        else:
                            # Create the locator
                            wait_operates.append(ovsdb.wait('Physical_Locator',
                                                            [["dst_ip", "==", ip],
                                                             ["encapsulation_type", "==", "vxlan_over_ipv4"]],
                                                            ["_uuid"],
                                                            [], True, 0))
                            name = 'locator_uuid_' + hex(ip4_addr(ip))
                            locator_uuid_dict[ip] = ovsdb.named_uuid(name)
                            set_operates.append(ovsdb.insert('Physical_Locator',
                                                             {"dst_ip": ip,
                                                              "encapsulation_type": "vxlan_over_ipv4"},
                                                             name))
                    if result[1]['rows'] and (not update or broadcasts_updated):
                        # locator set cannot be modified, it can only be replaced with a new set
                        mcast_uuid = result[1]['rows'][0]['locator_set']
                        # locator set is already created
                        if not broadcast_ips:
                            # Remove the whole locator set. We do not do check; it is always removed
                            set_operates.append(ovsdb.delete('Mcast_Macs_Remote',
                                                             [["MAC", "==", "unknown-dst"],
                                                              ["logical_switch", "==", ovsdb.uuid(lsuuid)]]))
                        else:
                            wait_operates.append(ovsdb.wait('Mcast_Macs_Remote',
                                                          [["MAC", "==", "unknown-dst"],
                                                          ["logical_switch", "==", ovsdb.uuid(lsuuid)]],
                                                          ["_uuid"],
                                                          [{"_uuid": result[1]['rows'][0]['_uuid']}],
                                                          True, 0))
                            # Create a new set
                            set_operates.append(ovsdb.insert('Physical_Locator_Set',
                                                             {"locators": ovsdb.oset(*[locator_uuid_dict[ip] for ip in broadcast_ips])},
                                                             'mcast_unknown_uuid'))
                            mcast_uuid = ovsdb.named_uuid('mcast_unknown_uuid')
                            set_operates.append(ovsdb.update('Mcast_Macs_Remote',
                                                            [["_uuid", "==", result[1]['rows'][0]['_uuid']]],
                                                             {"locator_set": mcast_uuid}))
                    else:
                        if broadcast_ips:
                            wait_operates.append(ovsdb.wait('Mcast_Macs_Remote',
                                                          [["MAC", "==", "unknown-dst"],
                                                          ["logical_switch", "==", ovsdb.uuid(lsuuid)]],
                                                          ["_uuid"],
                                                          [],
                                                          True, 0))
                            set_operates.append(ovsdb.insert('Physical_Locator_Set',
                                                             {"locators": ovsdb.oset(*[locator_uuid_dict[ip] for ip in broadcast_ips])},
                                                             'mcast_unknown_uuid'))
                            mcast_uuid = ovsdb.named_uuid('mcast_unknown_uuid')
                            set_operates.append(ovsdb.insert('Mcast_Macs_Remote',
                                                             {"MAC": "unknown-dst",
                                                              "logical_switch": ovsdb.uuid(lsuuid),
                                                              "locator_set": mcast_uuid}))
                    # modify unicast table
                    if update:
                        set_operates.extend(ovsdb.delete('Ucast_Macs_Remote',
                                                         [["logical_switch", "==", ovsdb.uuid(lsuuid)],
                                                          ["MAC", "==", mac]])
                                            for mac in remove_ucasts)
                        set_operates.extend(ovsdb.delete('Ucast_Macs_Remote',
                                                         [["logical_switch", "==", ovsdb.uuid(lsuuid)],
                                                          ["MAC", "==", mac]])
                                            for mac in add_ucasts)
                        set_operates.extend(ovsdb.insert('Ucast_Macs_Remote',
                                                         {"MAC": mac,
                                                          "logical_switch": ovsdb.uuid(lsuuid),
                                                          "locator": locator_uuid_dict[tunnel]})
                                            for mac, tunnel in add_ucasts.items())
                    else:
                        set_operates.append(ovsdb.delete('Ucast_Macs_Remote',
                                                         [["logical_switch", "==", ovsdb.uuid(lsuuid)]]))
                        set_operates.extend(ovsdb.insert('Ucast_Macs_Remote',
                                                         {"MAC": mac,
                                                          "logical_switch": ovsdb.uuid(lsuuid),
                                                          "locator": locator_uuid_dict[tunnel]})
                                            for mac, tunnel in add_ucasts.items())
                    method, params = ovsdb.transact('hardware_vtep', *(wait_operates + set_operates))
                    result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                    if any(r is None or 'error' in r for r in result[:len(wait_operates)]):
                        # Some wait operates failed, retry
                        continue
                    _check_transact_result(result[len(wait_operates):], set_operates)
                    break
            except Exception:
                self._logger.warning('Update hardware_vtep failed with exception', exc_info = True)
                # Force a full update next time
                _last_endpoints[:] = [set(), {}, None]
            else:
                _last_endpoints[:] = [broadcast_ips, mac_ip_dict, ps]
        try:
            while True:
                walk_result = await call_api(self.apiroutine, 'objectdb', 'walk',
                                 {'keys': (lognet_key,
                                           endpointset_key,
                                           lognetmap_key),
                                  'walkerdict': {lognet_key: walker},
                                  'requestid': requestid
                                  })
                if _update_set:
                    _update_set.clear()
                    continue
                lastkeys = set(_savedkeys)
                _savedkeys, _savedresult = walk_result
                if _detect_update_routine is not None:
                    _detect_update_routine.close()
                _detect_update_routine = self.apiroutine.subroutine(_detect_update(_savedresult), False)
                removekeys = tuple(lastkeys.difference(_savedkeys))
                if removekeys:
                    # Unwatch unnecessary keys
                    await call_api(self.apiroutine, 'objectdb', 'munwatch', {'keys': removekeys,
                                                                    'requestid': requestid})
                await _update_hardware_vtep(_savedresult)
                while True:
                    if not _update_set:
                        await update_matcher
                    should_rewalk = any(v.getkey() in rewalk_keys for v in _update_set if v is not None)
                    _update_set.clear()
                    if should_rewalk:
                        break
                    else:
                        await _update_hardware_vtep(_savedresult)
        finally:
            if _detect_update_routine is not None:
                _detect_update_routine.close()
            self.apiroutine.subroutine(send_api(self.apiroutine,'objectdb', 'unwatchall', {'requestid': requestid}), False)
    
    async def _update_logicalswitch_info(self, conn, logicalswitch, lsuuid):
        await self.apiroutine.execute_all([self._push_logicalswitch_endpoints(conn, logicalswitch),
                                           self._poll_logicalswitch_endpoints(conn, logicalswitch, lsuuid)])
    
    async def _manage_existing(self):
        conns = await call_api(self.apiroutine, "jsonrpcserver", "getconnections", {})
        vb = self.vhostbind
        matchers = []
        for c in conns:
            if vb is None or c.protocol.vhost in vb:
                self.apiroutine.subroutine(self._monitor_conn(c, True))
                matchers.append(VtepConnectionSynchronized.createMatcher(c))
        await self.apiroutine.wait_for_all(*matchers)
        self._synchronized = True
        await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'synchronized'))
            
    async def _recycling(self):
        interval = self.recycleinterval
        if interval is None:
            return
        async def _recycle_logical_switches(conn):
            try:
                protocol = conn.protocol
                method, params = ovsdb.transact('hardware_vtep',
                                                ovsdb.select('Logical_Switch',
                                                             [],
                                                             ["_uuid", "name"]))
                jsonrpc_result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                ls_list = [(r['_uuid'][1], r['name']) for r in jsonrpc_result[0]['rows']]
                for l, name in ls_list:
                    method, params = ovsdb.transact('hardware_vtep',
                                                    ovsdb.delete('Ucast_Macs_Remote',
                                                                 [["logical_switch", "==", ovsdb.uuid(l)]]),
                                                    ovsdb.delete('Mcast_Macs_Remote',
                                                                 [["logical_switch", "==", ovsdb.uuid(l)]]),
                                                    ovsdb.delete('Ucast_Macs_Local',
                                                                 [["logical_switch", "==", ovsdb.uuid(l)]]),
                                                    ovsdb.delete('Mcast_Macs_Local',
                                                                 [["logical_switch", "==", ovsdb.uuid(l)]]),
                                                    ovsdb.delete('Logical_Switch',
                                                                 [["_uuid", "==", ovsdb.uuid(l)]]))
                    jsonrpc_result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                    # If the logical switch is still referenced, the transact will fail
                    if 'error' in jsonrpc_result[0]:
                        self._logger.warning('Recycling logical switch %r (uuid = %r) failed: %r',
                                             name, l, jsonrpc_result[0]['error'])
                    elif not any(r for r in jsonrpc_result if 'error' in r):
                        self._logger.info('Recycle logical switch %r for no longer in use', name)
            except Exception:
                self._logger.warning('Recycling failed on connecton %r', conn, exc_info = True)
        while True:
            await self.apiroutine.wait_with_timeout(interval)
            await self.apiroutine.execute_all([_recycle_logical_switches(c)
                                               for c in self._connection_ps])

    async def _main(self):
        self.apiroutine.subroutine(self._manage_existing())
        self.apiroutine.subroutine(self._recycling(), True, '_recycling_routine')
        try:
            vb = self.vhostbind
            if vb is not None:
                conn_up = JsonRPCConnectionStateEvent.createMatcher(state = JsonRPCConnectionStateEvent.CONNECTION_UP,
                                                                     _ismatch = lambda x: x.createby.vhost in vb)
            else:
                conn_up = JsonRPCConnectionStateEvent.createMatcher(state = JsonRPCConnectionStateEvent.CONNECTION_UP)
            while True:
                ev = await conn_up
                self.apiroutine.subroutine(self._monitor_conn(ev.connection))
        finally:
            for r in list(self._monitor_routines):
                r.close()
            if hasattr(self.apiroutine, '_recycling_routine'):
                self.apiroutine._recycling_routine.close()
        
    async def listphysicalports(self, physicalswitch = None):
        '''
        Get physical ports list from this controller, grouped by physical switch name
        
        :param physicalswitch: physicalswitch name. Return all switches if is None.
        
        :return: dictionary: {physicalswitch: [physicalports]} e.g. {'ps1': ['port1', 'port2']}
        '''
        await self._wait_for_sync()
        async def _getports(conn):
            try:
                method, params = ovsdb.transact('hardware_vtep',
                                                ovsdb.select('Physical_Switch', [], ["name", "ports"]),
                                                ovsdb.select('Physical_Port', [], ["_uuid", "name"]))
                result, _ = await conn.protocol.querywithreply(method, params, conn, self.apiroutine)
                if 'error' in result[0] or 'error' in result[1]:
                    raise JsonRPCErrorResultException('OVSDB request failed: ' + repr(result))
                switches = result[0]['rows']
                ports = result[1]['rows']
                ports_dict = dict((r['_uuid'][1], r['name']) for r in ports)
                return dict((switch['name'],
                             [ports_dict[p[1]] for p in ovsdb.getlist(switch['ports'])
                              if p[1] in ports_dict])
                            for switch in switches)
            except Exception:
                self._logger.warning('Query OVSDB on %r failed with exception; will ignore this connection',
                                     conn, exc_info = True)
                return {}
        if physicalswitch is None:
            routines = [_getports(c) for c in self._connection_ps]
            result = await self.apiroutine.execute_all(routines)
            all_result = {}
            for r in result:
                all_result.update(r)
            return all_result
        else:
            if physicalswitch not in self._physical_switchs:
                return {}
            else:
                result = await _getports(self._physical_switchs[physicalswitch])
                if physicalswitch in result:
                    return {physicalswitch: result[physicalswitch]}
                else:
                    return {}
    
    async def listphysicalswitches(self, physicalswitch = None):
        '''
        Get physical switch info
        
        :param physicalswitch: physicalswitch name. Return all switches if is None.
        
        :return: dictionary: {physicalswitch: {key: value}} keys include: management_ips,
                             tunnel_ips, description, switch_fault_status
        '''
        await self._wait_for_sync()
        async def _getswitch(conn):
            try:
                method, params = ovsdb.transact('hardware_vtep',
                                                ovsdb.select('Physical_Switch', [],
                                                             ["name", "management_ips", "tunnel_ips",
                                                              "description", "switch_fault_status"]))
                result, _ = await conn.protocol.querywithreply(method, params, conn, self.apiroutine)
                if 'error' in result[0]:
                    raise JsonRPCErrorResultException('OVSDB request failed: ' + repr(result))
                switches = result[0]['rows']
                return dict((s['name'],
                             {'management_ips': ovsdb.getlist(s['management_ips']),
                              'tunnel_ips': ovsdb.getlist(s['tunnel_ips']),
                              'description': s['description'],
                              'switch_fault_status': ovsdb.getlist(s['switch_fault_status'])})
                            for s in switches)
            except Exception:
                self._logger.warning('Query OVSDB on %r failed with exception; will ignore this connection',
                                     conn, exc_info = True)
                return {}
        if physicalswitch is None:
            routines = [_getswitch(c) for c in self._connection_ps]
            result = await self.apiroutine.execute_all(routines)
            all_result = {}
            for r in result:
                all_result.update(r)
            return all_result
        else:
            if physicalswitch not in self._physical_switchs:
                return {}
            else:
                result = await _getswitch(self._physical_switchs[physicalswitch])
                if physicalswitch in result:
                    return {physicalswitch: result[physicalswitch]}
                else:
                    return {}

    async def updatelogicalswitch(self, physicalswitch, physicalport, vlanid, logicalnetwork, vni, logicalports):
        '''
        Bind VLAN on physicalport to specified logical network, and update logical port vxlan info
        
        :param physicalswitch: physical switch name, should be the name in PhysicalSwitch table of OVSDB vtep database
        
        :param physicalport: physical port name, should be the name in OVSDB vtep database
        
        :param vlanid: the vlan tag used for this logicalswitch
        
        :param logicalnetwork: the logical network id, will also be the logical switch id
        
        :param vni: the VXLAN VNI of the logical network
        
        :param logicalports: a list of logical port IDs. The VXLAN info of these ports will be updated.
        '''
        await self._wait_for_sync()
        vlanid = int(vlanid)
        vni = int(vni)
        if physicalswitch is None:
            raise ValueError('Physical switch cannot be None')
        # We may retry some times
        while True:
            if physicalswitch not in self._physical_switchs:
                timeout, _, _ = await self.apiroutine.wait_with_timeout(
                                                         5,
                                                         VtepPhysicalSwitchStateChanged.createMatcher(
                                                            VtepPhysicalSwitchStateChanged.UP,
                                                            physicalswitch))
                if timeout or physicalswitch not in self._physical_switchs:
                    raise ValueError('Physical switch %r is not found' % (physicalswitch,))
            conn, tunnelip = self._physical_switchs[physicalswitch]
            protocol = conn.protocol
            # First use a select to validate the request
            try:
                method, params = ovsdb.transact('hardware_vtep',
                                                ovsdb.select('Physical_Switch',
                                                             [["name", "==", physicalswitch]],
                                                             ["_uuid", "name", "ports"]),
                                                ovsdb.select('Physical_Port',
                                                             [["name", "==", physicalport]],
                                                             ["_uuid", "name", "vlan_bindings"]),
                                                ovsdb.select('Logical_Switch',
                                                             [["name", "==", logicalnetwork]],
                                                             ["_uuid", "name"])
                                                )
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                if 'rows' not in result[0]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[0]))
                if 'rows' not in result[1]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[1]))
                if 'rows' not in result[2]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[2]))
                if not result[0]['rows']:
                    raise ValueError('Physical switch %r is not found, connection = %r' % (physicalswitch, conn))
                pswitch = result[0]['rows'][0]
                port_uuids = set(p[1] for p in ovsdb.getlist(pswitch['ports']))
                ports = [p for p in result[1]['rows'] if p['_uuid'][1] in port_uuids]
                if not ports:
                    raise ValueError('Physical port %r is not found in physical switch %r' % (physicalport, physicalswitch))
                port = ports[0]
                ls_list = result[2]['rows']
                if ls_list:
                    ls_uuid = ls_list[0]['_uuid'][1]
                    # Logical switch is already created
                    curr_network = ovsdb.omap_getvalue(port['vlan_bindings'], vlanid)
                    if curr_network is None:
                        # Not binded
                        binded = False
                        created = True
                    elif curr_network[1] == ls_uuid:
                        # Already binded
                        binded = True
                        created = True
                    else:
                        raise ValueError('VLAN tag %r on physical port %r (%r) already binds to other networks, uuid=%r' \
                                         % (vlanid, physicalport, physicalswitch, curr_network[1]))
                else:
                    # Logical switch is not created
                    binded = False
                    created = False
                    curr_network = ovsdb.omap_getvalue(port['vlan_bindings'], vlanid)
                    if curr_network is not None:
                        raise ValueError('VLAN tag %r on physical port %r (%r) already binds to other networks, uuid=%r' \
                                         % (vlanid, physicalport, physicalswitch, curr_network[1]))
                if binded:
                    break
                operations = []
                # First we must check that the situations are not changed
                operations.extend([ovsdb.wait("Physical_Switch",
                                              [["name", "==", physicalswitch]],
                                              ["_uuid"],
                                              [{"_uuid": pswitch['_uuid']}],
                                              timeout = 0),
                                   ovsdb.wait("Physical_Port",
                                              [["name", "==", physicalport]],
                                              ["_uuid"],
                                              [{"_uuid": port['_uuid']}],
                                              timeout = 0)])
                if created:
                    operations.append(ovsdb.wait("Logical_Switch",
                                                [["name", "==", logicalnetwork]],
                                                ["_uuid"],
                                                [{"_uuid": ovsdb.uuid(ls_uuid)}],
                                                timeout = 0))
                    operations.append(ovsdb.update("Logical_Switch",
                                                   [["_uuid", "==", ovsdb.uuid(ls_uuid)]],
                                                   {"tunnel_key": vni}))
                else:
                    operations.append(ovsdb.wait("Logical_Switch",
                                                [["name", "==", logicalnetwork]],
                                                ["_uuid"],
                                                [],
                                                timeout = 0))
                    # Create the new logical switch
                    operations.append(ovsdb.insert('Logical_Switch',
                                                   {"name": logicalnetwork,
                                                    "tunnel_key": vni},
                                                   "new_logicalnetwork"))
                operations.append(ovsdb.mutate('Physical_Port', [["_uuid", "==", port['_uuid']]],
                                               [["vlan_bindings","insert",
                                                    ovsdb.omap(
                                                        (vlanid, ovsdb.uuid(ls_uuid) if created
                                                                 else ovsdb.named_uuid('new_logicalnetwork'))
                                                    )]]))
                method, params = ovsdb.transact('hardware_vtep', *operations)
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                if any(r for r in result[:3] if 'error' in r):
                    # Wait failed, the OVSDB is modified, retry
                    continue
                _check_transact_result(result[3:], operations[3:])
                break
            except ConnectionResetException:
                await self.apiroutine.wait_with_timeout(1)
                continue
            except IOError:
                await self.apiroutine.wait_with_timeout(1)
                continue
        if logicalports:
            # Refresh network with monitor, only update the logical port information
            await update_vxlaninfo(self.apiroutine,
                                      {},
                                      dict((p, tunnelip) for p in logicalports),
                                      {},
                                      protocol.vhost,
                                      physicalswitch,
                                      'hardware_vtep',
                                      self.allowedmigrationtime,
                                      self.refreshinterval)
        
    async def unbindlogicalswitch(self, physicalswitch, physicalport, vlanid, logicalnetwork):
        '''
        Remove bind of a physical port
        
        :param physicalswitch: physical switch name, should be the name in PhysicalSwitch table of OVSDB vtep database
        
        :param physicalport: physical port name, should be the name in OVSDB vtep database
        
        :param vlanid: the vlan tag used for this logicalswitch
        
        :param logicalnetwork: the logical network id, will also be the logical switch id
        '''
        await self._wait_for_sync()
        vlanid = int(vlanid)
        if physicalswitch is None:
            raise ValueError('Physical switch cannot be None')
        # We may retry some times
        while True:
            if physicalswitch not in self._physical_switchs:
                timeout, _, _ = await self.apiroutine.wait_with_timeout(
                                                    5,
                                                    VtepPhysicalSwitchStateChanged.createMatcher(
                                                        VtepPhysicalSwitchStateChanged.UP,
                                                        physicalswitch))
                if timeout or physicalswitch not in self._physical_switchs:
                    raise ValueError('Physical switch %r is not found' % (physicalswitch,))
            conn, tunnelip = self._physical_switchs[physicalswitch]
            protocol = conn.protocol
            try:
                method, params = ovsdb.transact('hardware_vtep',
                                                ovsdb.select('Physical_Switch',
                                                             [["name", "==", physicalswitch]],
                                                             ["_uuid", "name", "ports"]),
                                                ovsdb.select('Physical_Port',
                                                             [["name", "==", physicalport]],
                                                             ["_uuid", "name", "vlan_bindings"]),
                                                ovsdb.select('Logical_Switch',
                                                             [["name", "==", logicalnetwork]],
                                                             ["_uuid", "name"])
                                                )
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                if 'rows' not in result[0]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[0]))
                if 'rows' not in result[1]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[1]))
                if 'rows' not in result[2]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[2]))
                if not result[0]['rows']:
                    raise ValueError('Physical switch %r is not found, connection = %r' % (physicalswitch, conn))
                pswitch = result[0]['rows'][0]
                port_uuids = set(p[1] for p in ovsdb.getlist(pswitch['ports']))
                ports = [p for p in result[1]['rows'] if p['_uuid'][1] in port_uuids]
                if not ports:
                    raise ValueError('Physical port %r is not found in physical switch %r' % (physicalport, physicalswitch))
                port = ports[0]
                ls_list = result[2]['rows']
                if not ls_list:
                    # Network not exists, might already be unbinded, ignore
                    break
                ls_uuid = ls_list[0]['_uuid'][1]
                curr_network = ovsdb.omap_getvalue(port['vlan_bindings'], vlanid)
                if curr_network is None or curr_network[1] != ls_uuid:
                    # Not binded
                    break
                operations = []
                # First we must check that the situations are not changed
                operations.extend([ovsdb.wait("Physical_Switch",
                                              [["name", "==", physicalswitch]],
                                              ["_uuid"],
                                              [{"_uuid": pswitch['_uuid']}],
                                              timeout = 0),
                                   ovsdb.wait("Physical_Port",
                                              [["name", "==", physicalport],
                                               ["vlan_bindings", "includes", ovsdb.omap((vlanid, ovsdb.uuid(ls_uuid)))]],
                                              ["_uuid"],
                                              [{"_uuid": port['_uuid']}],
                                              timeout = 0),
                                   ovsdb.wait("Logical_Switch",
                                                [["name", "==", logicalnetwork]],
                                                ["_uuid"],
                                                [{"_uuid": ovsdb.uuid(ls_uuid)}],
                                                timeout = 0)])
                operations.append(ovsdb.mutate('Physical_Port', [["_uuid", "==", port['_uuid']]],
                                               [["vlan_bindings","delete",
                                                    ovsdb.omap(
                                                        (vlanid, ovsdb.uuid(ls_uuid))
                                                    )]]))
                # We do not delete the logical switch; it will be deleted by the recycling process if not used
                method, params = ovsdb.transact('hardware_vtep', *operations)
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                if any(r for r in result[:3] if 'error' in r):
                    # Wait failed, the OVSDB is modified, retry
                    continue
                _check_transact_result(result[3:], operations[3:])
                break
            except ConnectionResetException:
                await self.apiroutine.wait_with_timeout(1)
                continue
            except IOError:
                await self.apiroutine.wait_with_timeout(1)
                continue
    
    async def unbindphysicalport(self, physicalswitch, physicalport):
        '''
        Remove all bindings for a physical port

        :param physicalswitch: physical switch name, should be the name in PhysicalSwitch table of OVSDB vtep database
        
        :param physicalport: physical port name, should be the name in OVSDB vtep database        
        '''
        await self._wait_for_sync()
        if physicalswitch is None:
            raise ValueError('Physical switch cannot be None')
        # We may retry some times
        while True:
            if physicalswitch not in self._physical_switchs:
                timeout, _, _ = await self.apiroutine.waitWithTimeout(
                                            5,
                                            VtepPhysicalSwitchStateChanged.createMatcher(
                                                VtepPhysicalSwitchStateChanged.UP,
                                                physicalswitch))
                if timeout or physicalswitch not in self._physical_switchs:
                    raise ValueError('Physical switch %r is not found' % (physicalswitch,))
            conn, tunnelip = self._physical_switchs[physicalswitch]
            protocol = conn.protocol
            try:
                method, params = ovsdb.transact('hardware_vtep',
                                                ovsdb.select('Physical_Switch',
                                                             [["name", "==", physicalswitch]],
                                                             ["_uuid", "name", "ports"]),
                                                ovsdb.select('Physical_Port',
                                                             [["name", "==", physicalport]],
                                                             ["_uuid", "name", "vlan_bindings"])
                                                )
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                if 'rows' not in result[0]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[0]))
                if 'rows' not in result[1]:
                    raise JsonRPCErrorResultException('select from Physical_Switch failed: ' + repr(result[1]))
                if not result[0]['rows']:
                    raise ValueError('Physical switch %r is not found, connection = %r' % (physicalswitch, conn))
                pswitch = result[0]['rows'][0]
                port_uuids = set(p[1] for p in ovsdb.getlist(pswitch['ports']))
                ports = [p for p in result[1]['rows'] if p['_uuid'][1] in port_uuids]
                if not ports:
                    raise ValueError('Physical port %r is not found in physical switch %r' % (physicalport, physicalswitch))
                port = ports[0]
                # First we must check that the situations are not changed
                operations = [ovsdb.wait("Physical_Switch",
                                              [["name", "==", physicalswitch]],
                                              ["_uuid"],
                                              [{"_uuid": pswitch['_uuid']}],
                                              timeout = 0),
                               ovsdb.wait("Physical_Port",
                                          [["name", "==", physicalport]],
                                          ["_uuid"],
                                          [{"_uuid": port['_uuid']}],
                                          timeout = 0),
                               ovsdb.update("Physical_Port",
                                            [["_uuid", "==", port["_uuid"]]],
                                            {"vlan_bindings": ovsdb.omap()})]
                # We do not delete the logical switch; it will be deleted by the recycling process if not used
                method, params = ovsdb.transact('hardware_vtep', *operations)
                result, _ = await protocol.querywithreply(method, params, conn, self.apiroutine)
                if any(r for r in result[:2] if 'error' in r):
                    # Wait failed, the OVSDB is modified, retry
                    continue
                _check_transact_result(result[2:], operations[2:])
                break
            except ConnectionResetException:
                await self.apiroutine.wait_with_timeout(1)
                continue
            except IOError:
                await self.apiroutine.wait_with_timeout(1)
                continue



'''
Created on 2015/11/24

:author: hubo
'''
from vlcp.config import config
from vlcp.protocol.http import Http, HttpResponseEvent, HttpConnectionStateEvent,\
    HttpConnectionClosedException, HttpProtocolException, HttpResponseEndEvent
from email.message import Message
from vlcp.event.stream import MemoryStream
from vlcp.event.event import Event, withIndices, M_
from vlcp.event.connection import Client
from vlcp.event.core import TimerEvent
from vlcp.config.config import Configurable
from contextlib import closing

try:
    from Cookie import Morsel
    from cookielib import CookieJar, request_host
    from urlparse import urlsplit, urlunsplit, urljoin
    from urllib import quote
except Exception:
    from http.cookiejar import CookieJar, request_host
    from urllib.parse import urlsplit, urlunsplit, quote, urljoin
from vlcp.utils import encoders
import re
# From Python 2.7 ssl library

try:
    unicode
except Exception:
    unicode = str

def _dnsname_match(dn, hostname, max_wildcards=1):
    """Matching according to RFC 6125, section 6.4.3

    http://tools.ietf.org/html/rfc6125#section-6.4.3
    """
    pats = []
    if not dn:
        return False

    pieces = dn.split(r'.')
    leftmost = pieces[0]
    remainder = pieces[1:]

    wildcards = leftmost.count('*')
    if wildcards > max_wildcards:
        # Issue #17980: avoid denials of service by refusing more
        # than one wildcard per fragment.  A survery of established
        # policy among SSL implementations showed it to be a
        # reasonable choice.
        raise CertificateException(
            "too many wildcards in certificate DNS name: " + repr(dn))

    # speed up common case w/o wildcards
    if not wildcards:
        return dn.lower() == hostname.lower()

    # RFC 6125, section 6.4.3, subitem 1.
    # The client SHOULD NOT attempt to match a presented identifier in which
    # the wildcard character comprises a label other than the left-most label.
    if leftmost == '*':
        # When '*' is a fragment by itself, it matches a non-empty dotless
        # fragment.
        pats.append('[^.]+')
    elif leftmost.startswith('xn--') or hostname.startswith('xn--'):
        # RFC 6125, section 6.4.3, subitem 3.
        # The client SHOULD NOT attempt to match a presented identifier
        # where the wildcard character is embedded within an A-label or
        # U-label of an internationalized domain name.
        pats.append(re.escape(leftmost))
    else:
        # Otherwise, '*' matches any dotless string, e.g. www*
        pats.append(re.escape(leftmost).replace(r'\*', '[^.]*'))

    # add the remaining fragments, ignore any wildcards
    for frag in remainder:
        pats.append(re.escape(frag))

    pat = re.compile(r'\A' + r'\.'.join(pats) + r'\Z', re.IGNORECASE)
    return pat.match(hostname)


def match_hostname(cert, hostname):
    """Verify that *cert* (in decoded format as returned by
    SSLSocket.getpeercert()) matches the *hostname*.  RFC 2818 and RFC 6125
    rules are followed, but IP addresses are not accepted for *hostname*.

    CertificateError is raised on failure. On success, the function
    returns nothing.
    """
    if not cert:
        raise ValueError("empty or no certificate, match_hostname needs a "
                         "SSL socket or SSL context with either "
                         "CERT_OPTIONAL or CERT_REQUIRED")
    dnsnames = []
    san = cert.get('subjectAltName', ())
    for key, value in san:
        if key == 'DNS':
            if _dnsname_match(value, hostname):
                return
            dnsnames.append(value)
    if not dnsnames:
        # The subject is only checked when there is no dNSName entry
        # in subjectAltName
        for sub in cert.get('subject', ()):
            for key, value in sub:
                # XXX according to RFC 2818, the most specific Common Name
                # must be used.
                if key == 'commonName':
                    if _dnsname_match(value, hostname):
                        return
                    dnsnames.append(value)
    if len(dnsnames) > 1:
        raise CertificateException("hostname %r "
            "doesn't match either of %s"
            % (hostname, ', '.join(map(repr, dnsnames))))
    elif len(dnsnames) == 1:
        raise CertificateException("hostname %r "
            "doesn't match %r"
            % (hostname, dnsnames[0]))
    else:
        raise CertificateException("no appropriate commonName or "
            "subjectAltName fields were found")


def _str(s, encoding = 'ascii'):
    if not isinstance(s, str):
        return s.decode(encoding)
    else:
        return s


def _bytes(s, encoding = 'ascii'):
    if isinstance(s, bytes):
        return s
    else:
        return s.encode(encoding)


class WebException(IOError):
    pass


class CertificateException(IOError):
    pass


class ManualRedirectRequired(IOError):
    def __init__(self, msg, response, request, kwargs):
        IOError.__init__(self, msg)
        self.response = response
        self.location = response.get_header('Location')
        self.request = request
        self.kwargs = kwargs


class Request(object):
    def __init__(self, url, data = None, method = None, headers = {}, origin_req_host = None, unverifiable = False,
                 rawurl = False):
        '''
        :param url: request url
        :param data: request data, can be a str/bytes, or a stream(vlcp.event.stream.XXXStream)
        :param method: request method (GET, POST, ...)
        :param headers: request header dict ({'user-agent':'myagent'})
        :param origin_req_host: origin request host for cookie policy check
        :param unverifiable: unverifiable for cookie policy check
        '''
        self.url = _str(url, 'ascii')
        s = urlsplit(self.url, 'http')
        self.type = 'https' if s.scheme == 'https' else 'http'
        self.host = s.netloc
        if not self.host:
            raise ValueError('Invalid URL: ' + self.url)
        if rawurl:
            self.path = urlunsplit(('', '', s.path, s.query, ''))
        else:
            self.path = urlunsplit(('', '', quote(s.path), quote(s.query,'/&='), ''))
        if not self.path:
            self.path = '/'
        self.data = data
        if method is None:
            if self.data is None:
                self.method = 'GET'
            else:
                self.method = 'POST'
        else:
            self.method = method.upper()
        if self.data is not None:
            if isinstance(self.data, unicode):
                self.data = _bytes(self.data)
        headers = dict(headers)
        self.headers = dict((_str(k), _str(v, 'iso-8859-1')) for k,v in headers.items())
        self.headerdict = dict((k.lower(), v) for k,v in self.headers.items())
        self.headermap = dict((k.lower(), k) for k in self.headers.keys())
        self.undirectedheaders = {}
        self.undirectedheaderdict = {}
        self.undirectedheadermap = {}
        self.hostname = request_host(self)
        if origin_req_host is None:
            origin_req_host = request_host(self)
        self.origin_req_host = origin_req_host
        self.unverifiable = unverifiable
        self.redirect_count = 0
        if self.data and not self.has_header('Content-Type'):
            self.add_header('Content-Type', 'application/x-www-form-urlencoded')
    def get_full_url(self):
        return self.url
    def is_unverifiable(self):
        return self.unverifiable
    def get_type(self):
        return self.type
    def get_header(self, k, default = None):
        if k.lower() == 'host':
            return self.host
        else:
            return self.undirectedheaderdict.get(k.lower(), self.headerdict.get(k.lower(), default))
    def has_header(self, k):
        if k.lower() == 'host':
            return True
        else:
            return k.lower() in self.headerdict or k.lower() in self.undirectedheaderdict
    def get_origin_req_host(self):
        if self.origin_req_host is None:
            return request_host(self.host)
        else:
            return self.origin_req_host
    def header_items(self):
        d = dict(self.headerdict)
        d.update(self.undirectedheaderdict)
        return [(self.undirectedheadermap.get(k, self.headermap.get(k)),v) for k,v in d.items()]
    def add_header(self, k, v):
        self.headers[k] = v
        if k.lower() in self.headerdict:
            del self.headers[self.headermap[k.lower()]]
        self.headerdict[k.lower()] = v
        self.headermap[k.lower()] = k
    def add_unredirected_header(self, k, v):
        self.undirectedheaders[k] = v
        if k.lower() in self.undirectedheaderdict:
            del self.undirectedheaders[self.undirectedheadermap[k.lower()]]
        self.undirectedheaderdict[k.lower()] = v
        self.undirectedheadermap[k.lower()] = k
    def redirect(self, response, **kwargs):
        self.redirect_count += 1
        if self.redirect_count >= 16:
            raise WebException('Too many redirections')
        url = response.get_header('Location')
        if url is None:
            raise WebException('Receiving a 3xx response without Location header')
        oldurl = self.url
        self.url = url
        s = urlsplit(url, 'http')
        self.type = 'https' if s.scheme == 'https' else 'http'
        self.host = s.netloc
        if not self.host:
            # Redirect to a relative url
            self.url = urljoin(oldurl, url)
            s = urlsplit(self.url, 'http')
            self.host = s.netloc
        if not self.host:
            raise ValueError('Invalid redirect URL: ' + self.url)
        self.path = urlunsplit(('', '', quote(s.path), quote(s.query,'/&='), ''))
        self.undirectedheaderdict.clear()
        self.undirectedheadermap.clear()
        self.undirectedheaders.clear()
        if response.status in (307, 308):
            if self.data is not None and not isinstance(self.data, bytes):
                raise ManualRedirectRequired('Must retry post', response, self, kwargs)
        else:
            if self.method != 'HEAD':
                self.method = 'GET'
            self.data = None
class Response(object):
    def __init__(self, url, event, scheduler):
        self.url = url
        self.connection = event.connection
        self.iserror = event.iserror
        self.status = event.statuscode
        self.reason = _str(event.statustext)
        self.fullstatus = _str(event.status)
        self.code = self.status
        self.msg = self.reason
        self.version = event.connection.http_remoteversion
        self._headers = event.headers
        self._headerdict = event.headerdict
        self.headers = [(_str(k), _str(v, 'ISO-8859-1')) for k,v in event.headers]
        self.headerdict = dict((_str(k), _str(v, 'ISO-8859-1')) for k,v in event.headerdict.items())
        self._setcookies = event.setcookies
        self.setcookies = [_str(v, 'ISO-8859-1') for v in event.setcookies]
        self.stream = event.stream
        self.scheduler = scheduler
    def getcode(self):
        return self.code
    def geturl(self):
        return self.url
    def info(self):
        return self
    def get_header(self, k, default = None):
        return self.headerdict.get(k.lower(), default)
    def getheaders(self, k, default = []):
        if k.lower() == 'set-cookie':
            return self.setcookies
        elif k.lower() == 'set-cookie2':
            # Ignore these headers, since they are deprecated
            return default
        else:
            return [v for k2,v in self.headers if k2.lower() == k.lower()]
    def get_all(self, k, default = []):
        return self.getheaders(k, default)
    def has_header(self, k):
        return k.lower() in self.headerdict
    def close(self):
        "Stop the output stream, but further download will still perform"
        if self.stream:
            self.stream.close(self.scheduler)
            self.stream = None
    async def shutdown(self):
        "Force stop the output stream, if there are more data to download, shutdown the connection"
        if self.stream:
            if not self.stream.dataeof and not self.stream.dataerror:
                self.stream.close(self.scheduler)
                await self.connection.shutdown()
            else:
                self.stream.close(self.scheduler)
            self.stream = None

    def __del__(self):
        self.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exctype, excvalue, traceback):
        self.close()
        return False

@withIndices('host', 'path', 'https')
class WebClientRequestDoneEvent(Event):
    pass

@config('webclient')
class WebClient(Configurable):
    "Convenient HTTP request processing. Proxy is not supported in current version."
    # When a cleanup task is created, the task releases dead connections by this interval
    _default_cleanupinterval = 60
    # Persist this number of connections at most for each host. If all connections are in
    # use, new requests will wait until there are free connections.
    _default_samehostlimit = 20
    # Do not allow multiple requests to the same URL at the same time. If sameurllimit=True,
    # requests to the same URL will always be done sequential.
    _default_sameurllimit = False
    # CA file used to verify HTTPS certificates. To be compatible with older Python versions,
    # the new SSLContext is not enabled currently, so with the default configuration, the
    # certificates are NOT verified. You may configure this to a .pem file in your system,
    # usually /etc/pki/tls/cert.pem in Linux.
    _default_cafile = None
    # When following redirects and the server redirects too many times, raises an exception
    # and end the process
    _default_redirectlimit = 10
    # Verify the host with the host in certificate
    _default_verifyhost = True
    def __init__(self, allowcookies = False, cookiejar = None):
        '''
        :param allowcookies: Accept and store cookies, automatically use them on further requests
        :param cookiejar: Provide a customized cookiejar instead of the default CookieJar()
        '''
        self._connmap = {}
        self._requesting = set()
        self._hostwaiting = set()
        self._pathwaiting = set()
        self._protocol = Http(False)
        self.allowcookies = allowcookies
        if cookiejar is None:
            self.cookiejar = CookieJar()
        else:
            self.cookiejar = cookiejar
        self._tasks = []
    async def open(self, container, request, ignorewebexception = False, timeout = None, datagen = None, cafile = None, key = None, certificate = None,
             followredirect = True, autodecompress = False, allowcookies = None):
        '''
        Open http request with a Request object
        
        :param container: a routine container hosting this routine
        :param request: vlcp.utils.webclient.Request object
        :param ignorewebexception: Do not raise exception on Web errors (4xx, 5xx), return a response normally
        :param timeout: timeout on connection and single http request. When following redirect, new request
               does not share the old timeout, which means if timeout=2:
               connect to host: (2s)
               wait for response: (2s)
               response is 302, redirect
               connect to redirected host: (2s)
               wait for response: (2s)
               ...
               
        :param datagen: if the request use a stream as the data parameter, you may provide a routine to generate
                        data for the stream. If the request failed early, this routine is automatically terminated.
                        
        :param cafile: provide a CA file for SSL certification check. If not provided, the SSL connection is NOT verified.
        :param key: provide a key file, for client certification (usually not necessary)
        :param certificate: provide a certificate file, for client certification (usually not necessary)
        :param followredirect: if True (default), automatically follow 3xx redirections
        :param autodecompress: if True, automatically detect Content-Encoding header and decode the body
        :param allowcookies: override default settings to disable the cookies
        '''
        if cafile is None:
            cafile = self.cafile
        if allowcookies is None:
            allowcookies = self.allowcookies
        forcecreate = False
        datagen_routine = None
        if autodecompress:
            if not request.has_header('Accept-Encoding'):
                request.add_header('Accept-Encoding', 'gzip, deflate')
        while True:
            # Find or create a connection
            conn, created = await self._getconnection(container, request.host, request.path, request.get_type() == 'https',
                                                      forcecreate, cafile, key, certificate, timeout)
            # Send request on conn and wait for reply
            try:
                if allowcookies:
                    self.cookiejar.add_cookie_header(request)
                if isinstance(request.data, bytes):
                    stream = MemoryStream(request.data)
                else:
                    stream = request.data
                if datagen and datagen_routine is None:
                    datagen_routine = container.subroutine(datagen)
                else:
                    datagen_routine = None
                timeout_, result = await container.execute_with_timeout(
                                                timeout,
                                                self._protocol.request_with_response(
                                                        container,
                                                        conn,
                                                        _bytes(request.host),
                                                        _bytes(request.path),
                                                        _bytes(request.method),
                                                        [(_bytes(k), _bytes(v)) for k,v in request.header_items()], stream))
                if timeout_:
                    if datagen_routine:
                        container.terminate(datagen_routine)
                    container.subroutine(self._releaseconnection(conn, request.host, request.path, request.get_type() == 'https', True), False)
                    raise WebException('HTTP request timeout')
                finalresp, _ = result
                resp = Response(request.get_full_url(), finalresp, container.scheduler)
                if allowcookies:
                    self.cookiejar.extract_cookies(resp, request)
                if resp.iserror and not ignorewebexception:
                    try:
                        exc = WebException(resp.fullstatus)
                        if autodecompress and resp.stream:
                            ce = resp.get_header('Content-Encoding', '')
                            if ce.lower() == 'gzip' or ce.lower() == 'x-gzip':
                                resp.stream.getEncoderList().append(encoders.gzip_decoder())
                            elif ce.lower() == 'deflate':
                                resp.stream.getEncoderList().append(encoders.deflate_decoder())
                        data = await resp.stream.read(container, 4096)
                        exc.response = resp
                        exc.body = data
                        if datagen_routine:
                            container.terminate(datagen_routine)
                        await resp.shutdown()
                        container.subroutine(self._releaseconnection(conn, request.host, request.path, request.get_type() == 'https', True), False)
                        raise exc
                    finally:
                        resp.close()
                else:
                    try:
                        container.subroutine(self._releaseconnection(conn, request.host, request.path, request.get_type() == 'https', False, finalresp), False)
                        if followredirect and resp.status in (300, 301, 302, 303, 307, 308):
                            request.redirect(resp, ignorewebexception = ignorewebexception, timeout = timeout, cafile = cafile, key = key,
                                             certificate = certificate, followredirect = followredirect,
                                             autodecompress = autodecompress, allowcookies = allowcookies)
                            resp.close()
                            continue
                        if autodecompress and resp.stream:
                            ce = resp.get_header('Content-Encoding', '')
                            if ce.lower() == 'gzip' or ce.lower() == 'x-gzip':
                                resp.stream.getEncoderList().append(encoders.gzip_decoder())
                            elif ce.lower() == 'deflate':
                                resp.stream.getEncoderList().append(encoders.deflate_decoder())
                        return resp
                    except:
                        resp.close()
                        raise
            except HttpConnectionClosedException:
                await self._releaseconnection(conn, request.host, request.path, request.get_type() == 'https', False)
                if not created:
                    # Retry on a newly created connection
                    forcecreate = True
                    continue
                else:
                    if datagen_routine:
                        container.terminate(datagen_routine)
                    raise
            except Exception as exc:
                await self._releaseconnection(conn, request.host, request.path, request.get_type() == 'https', True)
                raise exc
            break
    async def _releaseconnection(self, connection, host, path, https = False, forceclose = False, respevent = None):
        if not host:
            raise ValueError
        if forceclose:
            await connection.shutdown(True)
        if not forceclose and connection.connected and respevent:
            async def releaseconn():
                keepalive = await self._protocol.wait_for_response_end(connection, connection, respevent.connmark, respevent.xid)
                conns = self._connmap[host]
                conns[2] -= 1
                if keepalive:
                    connection.setdaemon(True)
                    conns[1 if https else 0].append(connection)
                else:
                    await connection.shutdown()
            connection.subroutine(releaseconn(), False)
        else:
            conns = self._connmap[host]
            conns[2] -= 1
        if self.sameurllimit:
            self._requesting.remove((host, path, https))
        if (host, path, https) in self._pathwaiting or host in self._hostwaiting:
            await connection.wait_for_send(WebClientRequestDoneEvent(host, path, https))
            if (host, path, https) in self._pathwaiting:
                self._pathwaiting.remove((host, path, https))
            if host in self._hostwaiting:
                self._hostwaiting.remove(host)
    async def _getconnection(self, container, host, path, https = False, forcecreate = False, cafile = None, key = None, certificate = None,
                       timeout = None):
        if not host:
            raise ValueError
        matcher = WebClientRequestDoneEvent.createMatcher(host, path, https)
        while self.sameurllimit and (host, path, https) in self._requesting:
            self._pathwaiting.add((host, path, https))
            await matcher
        # Lock the path
        if self.sameurllimit:
            self._requesting.add((host, path, https))
        # connmap format: (free, free_ssl, workingcount)
        conns = self._connmap.setdefault(host, [[],[], 0])
        conns[0] = [c for c in conns[0] if c.connected]
        conns[1] = [c for c in conns[1] if c.connected]
        myset = conns[1 if https else 0]
        if not forcecreate and myset:
            # There are free connections, reuse them
            conn = myset.pop()
            conn.setdaemon(False)
            conns[2] += 1
            return (conn, False)
        matcher = WebClientRequestDoneEvent.createMatcher(host)
        while self.samehostlimit and len(conns[0]) + len(conns[1]) + conns[2] >= self.samehostlimit:
            if myset:
                # Close a old connection
                conn = myset.pop()
                await conn.shutdown()
            else:
                # Wait for free connections
                self._hostwaiting.add(host)
                await matcher
                conns = self._connmap.setdefault(host, [[],[], 0])
                myset = conns[1 if https else 0]
                if not forcecreate and myset:
                    conn = myset.pop()
                    conn.setdaemon(False)
                    conns[2] += 1
                    return (conn, False)
        # Create new connection
        conns[2] += 1
        conn = Client(urlunsplit(('ssl' if https else 'tcp', host, '/', '', '')), self._protocol, container.scheduler,
                      key, certificate, cafile)
        if timeout is not None:
            conn.connect_timeout = timeout
        conn.start()
        connected = self._protocol.statematcher(conn, HttpConnectionStateEvent.CLIENT_CONNECTED, False)
        notconnected = self._protocol.statematcher(conn, HttpConnectionStateEvent.CLIENT_NOTCONNECTED, False)
        _, m = await M_(connected, notconnected)
        if m is notconnected:
            conns[2] -= 1
            await conn.shutdown(True)
            raise IOError('Failed to connect to %r' % (conn.rawurl,))
        if https and cafile and self.verifyhost:
            try:
                # TODO: check with SSLContext
                hostcheck = re.sub(r':\d+$', '', host)
                if host == conn.socket.remoteaddr[0]:
                    # IP Address is currently now allowed
                    await conn.shutdown(True)
                    raise CertificateException('Cannot verify host with IP address')
                match_hostname(conn.socket.getpeercert(False), hostcheck)
            except:
                conns[2] -= 1
                raise
        return (conn, True)
    def cleanup(self, host = None):
        "Cleaning disconnected connections"
        if host is not None:
            conns = self._connmap.get(host)
            if conns is None:
                return
            # cleanup disconnected connections
            conns[0] = [c for c in conns[0] if c.connected]
            conns[1] = [c for c in conns[1] if c.connected]
            if not conns[0] and not conns[1] and not conns[2]:
                del self._connmap[host]
        else:
            hosts = list(self._connmap.keys())
            for h in hosts:
                self.cleanup(h)
    def cleanup_task(self, container, interval = None):
        '''
        If this client object is persist for a long time, and you are worrying about memory leak,
        create a routine with this method: myclient.cleanup_task(mycontainer, 60).
        But remember that if you have created at lease one task, you must call myclient.endtask()
        to completely release the webclient object.
        '''
        if interval is None:
            interval = self.cleanupinterval
        async def task():
            th = container.scheduler.setTimer(interval, interval)
            tm = TimerEvent.createMatcher(th)
            try:
                while True:
                    await tm
                    self.cleanup()
            finally:
                container.scheduler.cancelTimer(th)
        t = container.subroutine(task(), False, daemon = True)
        self._tasks.append(t)
        return t
    async def shutdown(self):
        "Shutdown free connections to release resources"
        for c0, c1, _ in list(self._connmap.values()):
            c0bak = list(c0)
            del c0[:]
            for c in c0bak:
                if c.connected:
                    await c.shutdown()
            c1bak = list(c1)
            del c1[:]
            for c in c1bak:
                if c.connected:
                    await c.shutdown()
    def endtask(self):
        for t in self._tasks:
            t.close()
        del self._tasks[:]
                
    async def urlopen(self, container, url, data = None, method = None, headers = {}, rawurl = False, *args, **kwargs):
        '''
        Similar to urllib2.urlopen, but:
        1. is a routine
        2. data can be an instance of vlcp.event.stream.BaseStream, or str/bytes
        3. can specify method
        4. if datagen is not None, it is a routine which writes to <data>. It is automatically terminated if the connection is down.
        5. can also specify key and certificate, for client certification
        6. certificates are verified with CA if provided.
        If there are keep-alived connections, they are automatically reused.
        See open for available arguments
        
        Extra argument:
        
        :param rawurl: if True, assume the url is already url-encoded, do not encode it again.
        '''
        return await self.open(container, Request(url, data, method, headers, rawurl=rawurl), *args, **kwargs)
    async def manualredirect(self, container, exc, data, datagen = None):
        "If data is a stream, it cannot be used again on redirect. Catch the ManualRedirectException and call a manual redirect with a new stream."
        request = exc.request
        request.data = data
        return await self.open(container, request, datagen = datagen, **exc.kwargs)
    async def urlgetcontent(self, container, url, data = None, method = None, headers = {}, tostr = False,  encoding = None, rawurl = False, *args, **kwargs):
        '''
        In Python2, bytes = str, so tostr and encoding has no effect.
        In Python3, bytes are decoded into unicode str with encoding.
        If encoding is not specified, charset in content-type is used if present, or default to utf-8 if not.
        See open for available arguments

        :param rawurl: if True, assume the url is already url-encoded, do not encode it again.
        '''
        req = Request(url, data, method, headers, rawurl = rawurl)
        with (await self.open(container, req, *args, **kwargs)) as resp:
            encoding = 'utf-8'
            if encoding is None:
                m = Message()
                m.add_header('Content-Type', resp.get_header('Content-Type', 'text/html'))
                encoding = m.get_content_charset('utf-8')
            if not resp.stream:
                content = b''
            else:
                content = await resp.stream.read(container)
            if tostr:
                content = _str(content, encoding)
            return content
